import runpy
import sys
from io import StringIO
from time import sleep

import importlib_resources
import pkg_resources
from benchmarker import Benchmarker
from importlib_resources import path

from benchmarks.benchmark_utils import swap_attr

for entry in [entry for entry in importlib_resources.contents('examples') if
              not pkg_resources.resource_isdir('examples', entry)
              and entry.endswith(".py")
              and '__init__' not in entry
              and 'demo.py' not in entry
              ]:
    print("*****************************************************************")
    print("Benchmarking {}".format(entry))
    print("*****************************************************************")

    with path('examples', entry) as file_in:
        with Benchmarker(cycle=20, extra=1) as bench:
            @bench(entry)
            def _(_):
                # prevent Benchmarker from doing "ZeroDivisionError: float division by zero:
                #    ratio = base_time / real_time"
                sleep(0.001)
                # In order to pipe input into examples that have input(),
                # we use the test package, which is meant for internal use by Python only internal and
                # Any use of this package outside of Python’s standard library is discouraged as code (..)
                # can change or be removed without notice between releases of Python.
                # https://docs.python.org/3/library/test.html
                # TODO remove input() from examples?
                with swap_attr(sys, "stdin", StringIO("q")):
                    runpy.run_path(file_in, run_name="__main__")



import os
import time
from collections import deque

import ipywidgets
import jpy_canvas
import numpy as np
from ipywidgets import IntSlider, VBox, HBox, Checkbox, Output, Text, RadioButtons, Tab
from numpy import array

import flatland.utils.rendertools as rt
from flatland.core.grid.grid4_utils import mirror
from flatland.envs.agent_utils import EnvAgent, EnvAgentStatic
from flatland.envs.generators import complex_rail_generator, empty_rail_generator
from flatland.envs.observations import TreeObsForRailEnv
from flatland.envs.rail_env import RailEnv, random_rail_generator


class EditorMVC(object):
    """ EditorMVC - a class to encompass and assemble the Jupyter Editor Model-View-Controller.
    """

    def __init__(self, env=None, sGL="PIL"):
        """ Create an Editor MVC assembly around a railenv, or create one if None.
        """
        if env is None:
            env = RailEnv(width=10,
                          height=10,
                          rail_generator=empty_rail_generator(),
                          number_of_agents=0,
                          obs_builder_object=TreeObsForRailEnv(max_depth=2))

        env.reset()

        self.editor = EditorModel(env)
        self.editor.view = self.view = View(self.editor, sGL=sGL)
        self.view.controller = self.editor.controller = self.controller = Controller(self.editor, self.view)
        self.view.init_canvas()
        self.view.init_widgets()  # has to be done after controller


class View(object):
    """ The Jupyter Editor View - creates and holds the widgets comprising the Editor.
    """

    def __init__(self, editor, sGL="MPL"):
        self.editor = self.model = editor
        self.sGL = sGL

    def display(self):
        self.output_generator.clear_output()
        return self.wMain

    def init_canvas(self):
        # update the rendertool with the env
        self.new_env()
        self.oRT.render_env(show=False)
        img = self.oRT.get_image()
        self.wImage = jpy_canvas.Canvas(img)
        self.yxSize = self.wImage.data.shape[:2]
        self.writableData = np.copy(self.wImage.data)  # writable copy of image - wid_img.data is somehow readonly
        self.wImage.register_move(self.controller.on_mouse_move)
        self.wImage.register_click(self.controller.on_click)

        self.yxBase = self.oRT.gl.yxBase
        self.nPixCell = self.oRT.gl.nPixCell

    def init_widgets(self):
        # Debug checkbox - enable logging in the Output widget
        self.debug = ipywidgets.Checkbox(description="Debug")
        self.debug.observe(self.controller.set_debug, names="value")

        # Separate checkbox for mouse move events - they are very verbose
        self.debug_move = Checkbox(description="Debug mouse move")
        self.debug_move.observe(self.controller.set_debug_move, names="value")

        # This is like a cell widget where loggin goes
        self.output_generator = Output()

        # Filename textbox
        self.filename = Text(description="Filename")
        self.filename.value = self.model.env_filename
        self.filename.observe(self.controller.set_filename, names="value")

        # Size of environment when regenerating

        self.regen_width = IntSlider(value=10, min=5, max=100, step=5, description="Regen Size (Width)",
                                     tip="Click Regenerate after changing this")
        self.regen_width.observe(self.controller.set_regen_width, names="value")

        self.regen_height = IntSlider(value=10, min=5, max=100, step=5, description="Regen Size (Height)",
                                      tip="Click Regenerate after changing this")
        self.regen_height.observe(self.controller.set_regen_height, names="value")

        # Number of Agents when regenerating
        self.regen_n_agents = IntSlider(value=1, min=0, max=5, step=1, description="# Agents",
                                        tip="Click regenerate or reset after changing this")
        self.regen_method = RadioButtons(description="Regen\nMethod", options=["Empty", "Random Cell"])

        self.replace_agents = Checkbox(value=True, description="Replace Agents")

        self.wTab = Tab()
        tab_contents = ["Regen", "Observation"]
        for i, title in enumerate(tab_contents):
            self.wTab.set_title(i, title)
        self.wTab.children = [
            VBox([self.regen_width, self.regen_height, self.regen_n_agents, self.regen_method])
        ]

        # abbreviated description of buttons and the methods they call
        ldButtons = [
            dict(name="Refresh", method=self.controller.refresh, tip="Redraw only"),
            dict(name="Rotate Agent", method=self.controller.rotate_agent, tip="Rotate selected agent"),
            dict(name="Restart Agents", method=self.controller.restart_agents,
                 tip="Move agents back to start positions"),
            dict(name="Random", method=self.controller.reset,
                 tip="Generate a randomized scene, including regen rail + agents"),
            dict(name="Regenerate", method=self.controller.regenerate,
                 tip="Regenerate the rails using the method selected below"),
            dict(name="Load", method=self.controller.load),
            dict(name="Save", method=self.controller.save),
            dict(name="Save as image", method=self.controller.save_image)
        ]

        self.lwButtons = []
        for dButton in ldButtons:
            wButton = ipywidgets.Button(description=dButton["name"],
                                        tooltip=dButton["tip"] if "tip" in dButton else dButton["name"])
            wButton.on_click(dButton["method"])
            self.lwButtons.append(wButton)

        self.wVbox_controls = VBox([
            self.filename,
            *self.lwButtons,
            self.wTab])

        self.wMain = HBox([self.wImage, self.wVbox_controls])

    def draw_stroke(self):
        pass

    def new_env(self):
        """ Tell the view to update its graphics when a new env is created.
        """
        self.oRT = rt.RenderTool(self.editor.env, gl=self.sGL)

    def redraw(self):
        with self.output_generator:
            self.oRT.set_new_rail()
            self.model.env.agents = self.model.env.agents_static
            for a in self.model.env.agents:
                if hasattr(a, 'old_position') is False:
                    a.old_position = a.position
                if hasattr(a, 'old_direction') is False:
                    a.old_direction = a.direction

            self.oRT.render_env(agents=True,
                                show=False,
                                selected_agent=self.model.selected_agent,
                                show_observations=False)
            img = self.oRT.get_image()

            self.wImage.data = img
            self.writableData = np.copy(self.wImage.data)

            # the size should only be updated on regenerate at most
            self.yxSize = self.wImage.data.shape[:2]
            return img

    def redisplay_image(self):
        if self.writableData is not None:
            # This updates the image in the browser to be the new edited version
            self.wImage.data = self.writableData

    def drag_path_element(self, x, y):
        # Draw a black square on the in-memory copy of the image
        if x > 10 and x < self.yxSize[1] and y > 10 and y < self.yxSize[0]:
            self.writableData[(y - 2):(y + 2), (x - 2):(x + 2), :3] = 0

    def xy_to_rc(self, x, y):
        rcCell = ((array([y, x]) - self.yxBase))
        nX = np.floor((self.yxSize[0] - self.yxBase[0]) / self.model.env.height)
        nY = np.floor((self.yxSize[1] - self.yxBase[1]) / self.model.env.width)
        rcCell[0] = max(0, min(np.floor(rcCell[0] / nY), self.model.env.height - 1))
        rcCell[1] = max(0, min(np.floor(rcCell[1] / nX), self.model.env.width - 1))
        return rcCell

    def log(self, *args, **kwargs):
        if self.output_generator:
            with self.output_generator:
                print(*args, **kwargs)
        else:
            print(*args, **kwargs)


class Controller(object):
    """
    Controller to handle incoming events from the ipywidgets
    Updates the editor/model.
    Calls the View directly for things which do not directly effect the model
    (this means the mouse drag path before it is interpreted as transitions)
    """

    def __init__(self, model, view):
        self.editor = self.model = model
        self.view = view
        self.q_events = deque()
        self.drawMode = "Draw"

    def set_model(self, model):
        self.model = model

    def on_click(self, wid, event):
        x = event['canvasX']
        y = event['canvasY']
        self.debug("debug:", x, y)

        rcCell = self.view.xy_to_rc(x, y)

        bShift = event["shiftKey"]
        bCtrl = event["ctrlKey"]
        bAlt = event["altKey"]
        if bCtrl and not bShift and not bAlt:
            self.model.click_agent(rcCell)
            self.lrcStroke = []
        elif bShift and bCtrl:
            self.model.add_target(rcCell)
            self.lrcStroke = []
        elif bAlt and not bShift and not bCtrl:
            self.model.clear_cell(rcCell)
            self.lrcStroke = []

        self.debug("click in cell", rcCell)
        self.model.debug_cell(rcCell)

        if self.model.selected_agent is not None:
            self.lrcStroke = []

    def set_debug(self, event):
        self.model.set_debug(event["new"])

    def set_debug_move(self, event):
        self.model.set_debug_move(event["new"])

    def set_draw_mode(self, event):
        self.set_draw_mode = event["new"]

    def set_filename(self, event):
        self.model.set_filename(event["new"])

    def on_mouse_move(self, wid, event):
        """Mouse motion event handler for drawing.
        """

        x = event['canvasX']
        y = event['canvasY']
        q_events = self.q_events

        if self.model.debug_bool and (event["buttons"] > 0 or self.model.debug_move_bool):
            self.debug("debug:", len(q_events), event)

        # If the mouse is held down, enqueue an event in our own queue
        # The intention was to avoid too many redraws.
        # Reset the lrcStroke list, if ALT, CTRL or SHIFT pressed
        if event["buttons"] > 0:
            q_events.append((time.time(), x, y))
            bShift = event["shiftKey"]
            bCtrl = event["ctrlKey"]
            bAlt = event["altKey"]
            if bShift:
                self.lrcStroke = []
                while len(q_events) > 0:
                    t, x, y = q_events.popleft()
                return
            if bCtrl:
                self.lrcStroke = []
                while len(q_events) > 0:
                    t, x, y = q_events.popleft()
                return
            if bAlt:
                self.lrcStroke = []
                while len(q_events) > 0:
                    t, x, y = q_events.popleft()
                return
        else:
            self.lrcStroke = []

        if self.model.selected_agent is not None:
            self.lrcStroke = []
            while len(q_events) > 0:
                t, x, y = q_events.popleft()
            return

        # Process the events in our queue:
        # Draw a black square to indicate a trail
        # Convert the xy position to a cell rc
        # Enqueue transitions across cells in another queue
        if len(q_events) > 0:
            t_now = time.time()
            if t_now - q_events[0][0] > 0.1:  # wait before trying to draw

                while len(q_events) > 0:
                    t, x, y = q_events.popleft()  # get events from our queue
                    self.view.drag_path_element(x, y)

                    # Translate and scale from x,y to integer row,col (note order change)
                    rcCell = self.view.xy_to_rc(x, y)
                    self.editor.drag_path_element(rcCell)

                self.view.redisplay_image()

        else:
            self.model.mod_path(not event["shiftKey"])

    def refresh(self, event):
        self.debug("refresh")
        self.view.redraw()

    def clear(self, event):
        self.model.clear()

    def reset(self, event):
        self.log("Reset - nAgents:", self.view.regen_n_agents.value)
        self.log("Reset - size:", self.model.regen_size_width)
        self.log("Reset - size:", self.model.regen_size_height)
        self.model.reset(replace_agents=self.view.replace_agents.value,
                         nAgents=self.view.regen_n_agents.value)

    def rotate_agent(self, event):
        self.log("Rotate Agent:", self.model.selected_agent)
        if self.model.selected_agent is not None:
            for agent_idx, agent in enumerate(self.model.env.agents_static):
                if agent is None:
                    continue
                if agent_idx == self.model.selected_agent:
                    agent.direction = (agent.direction + 1) % 4
                    agent.old_direction = agent.direction
        self.model.redraw()

    def restart_agents(self, event):
        self.log("Restart Agents - nAgents:", self.view.regen_n_agents.value)
        if self.model.init_agents_static is not None:
            self.model.env.agents_static = [EnvAgentStatic(d[0], d[1], d[2], moving=False) for d in
                                            self.model.init_agents_static]
            self.model.env.agents = None
            self.model.init_agents_static = None
            self.model.env.restart_agents()
            self.model.env.reset(False, False)
        self.refresh(event)

    def regenerate(self, event):
        method = self.view.regen_method.value
        n_agents = self.view.regen_n_agents.value
        self.model.regenerate(method, n_agents)

    def set_regen_width(self, event):
        self.model.set_regen_width(event["new"])

    def set_regen_height(self, event):
        self.model.set_regen_height(event["new"])

    def load(self, event):
        self.model.load()

    def save(self, event):
        self.model.save()

    def save_image(self, event):
        self.model.save_image()

    def step(self, event):
        self.model.step()

    def log(self, *args, **kwargs):
        if self.view is None:
            print(*args, **kwargs)
        else:
            self.view.log(*args, **kwargs)

    def debug(self, *args, **kwargs):
        self.model.debug(*args, **kwargs)


class EditorModel(object):
    def __init__(self, env):
        self.view = None
        self.env = env
        self.regen_size_width = 10
        self.regen_size_height = 10

        self.lrcStroke = []
        self.iTransLast = -1
        self.gRCTrans = array([[-1, 0], [0, 1], [1, 0], [0, -1]])  # NESW in RC

        self.debug_bool = False
        self.debug_move_bool = False
        self.wid_output = None
        self.draw_mode = "Draw"
        self.env_filename = "temp.pkl"
        self.set_env(env)
        self.selected_agent = None
        self.init_agents_static = None
        self.thread = None
        self.save_image_count = 0

    def set_env(self, env):
        """
        set a new env for the editor, used by load and regenerate.
        """
        self.env = env

    def set_debug(self, debug):
        self.debug_bool = debug
        self.log("Set Debug:", self.debug_bool)

    def set_debug_move(self, debug):
        self.debug_move_bool = debug
        self.log("Set DebugMove:", self.debug_move_bool)

    def set_draw_mode(self, draw_mode):
        self.draw_mode = draw_mode

    def interpolate_path(self, rcLast, rcCell):
        if np.array_equal(rcLast, rcCell):
            return []
        rcLast = array(rcLast)
        rcCell = array(rcCell)
        rcDelta = rcCell - rcLast

        lrcInterp = []  # extra row,col points

        if np.any(np.abs(rcDelta) >= 1):
            iDim0 = np.argmax(np.abs(rcDelta))  # the dimension with the bigger move
            iDim1 = 1 - iDim0  # the dim with the smaller move
            rcRatio = rcDelta[iDim1] / rcDelta[iDim0]
            delta0 = rcDelta[iDim0]
            sgn0 = np.sign(delta0)

            iDelta1 = 0

            # count integers along the larger dimension
            for iDelta0 in range(sgn0, delta0 + sgn0, sgn0):
                rDelta1 = iDelta0 * rcRatio

                if np.abs(rDelta1 - iDelta1) >= 1:
                    rcInterp = (iDelta0, iDelta1)  # fill in the "corner" for "Manhattan interpolation"
                    lrcInterp.append(rcInterp)
                    iDelta1 = int(rDelta1)

                rcInterp = (iDelta0, int(rDelta1))
                lrcInterp.append(rcInterp)
            g2Interp = array(lrcInterp)
            if iDim0 == 1:  # if necessary, swap c,r to make r,c
                g2Interp = g2Interp[:, [1, 0]]
            g2Interp += rcLast
            # Convert the array to a list of tuples
            lrcInterp = list(map(tuple, g2Interp))
        return lrcInterp

    def drag_path_element(self, rcCell):
        """Mouse motion event handler for drawing.
        """
        lrcStroke = self.lrcStroke

        # Store the row,col location of the click, if we have entered a new cell
        if len(lrcStroke) > 0:
            rcLast = lrcStroke[-1]
            if not np.array_equal(rcLast, rcCell):  # only save at transition
                lrcInterp = self.interpolate_path(rcLast, rcCell)
                lrcStroke.extend(lrcInterp)
                self.debug("lrcStroke ", len(lrcStroke), rcCell, "interp:", lrcInterp)

        else:
            # This is the first cell in a mouse stroke
            lrcStroke.append(rcCell)
            self.debug("lrcStroke ", len(lrcStroke), rcCell)

    def mod_path(self, bAddRemove):
        # disabled functionality (no longer required)
        if bAddRemove is False:
            return
        # This elif means we wait until all the mouse events have been processed (black square drawn)
        # before trying to draw rails.  (We could change this behaviour)
        # Equivalent to waiting for mouse button to be lifted (and a mouse event is necessary:
        # the mouse may need to be moved)
        lrcStroke = self.lrcStroke
        if len(lrcStroke) >= 2:
            self.mod_rail_cell_seq(lrcStroke, bAddRemove)
            self.redraw()

    def mod_rail_cell_seq(self, lrcStroke, bAddRemove=True):
        # If we have already touched 3 cells
        # We have a transition into a cell, and out of it.

        if len(lrcStroke) >= 2:
            # If the first cell in a stroke is empty, add a deadend to cell 0
            if self.env.rail.get_full_transitions(*lrcStroke[0]) == 0:
                self.mod_rail_2cells(lrcStroke, bAddRemove, iCellToMod=0)

        # Add transitions for groups of 3 cells
        # hence inbound and outbound transitions for middle cell
        while len(lrcStroke) >= 3:
            self.mod_rail_3cells(lrcStroke, bAddRemove=bAddRemove)

        # If final cell empty, insert deadend:
        if len(lrcStroke) == 2:
            if self.env.rail.get_full_transitions(*lrcStroke[1]) == 0:
                self.mod_rail_2cells(lrcStroke, bAddRemove, iCellToMod=1)

        # now empty out the final two cells from the queue
        lrcStroke.clear()

    def mod_rail_3cells(self, lrcStroke, bAddRemove=True, bPop=True):
        """
        Add transitions for rail spanning three cells.
        lrcStroke -- list containing "stroke" of cells across grid
        bAddRemove -- whether to add (True) or remove (False) the transition
        The transition is added to or removed from the 2nd cell, consistent with
        entering from the 1st cell, and exiting into the 3rd.
        Both the forward and backward transitions are added,
        eg rcCells [(3,4), (2,4), (2,5)] would result in the transitions
        N->E and W->S in cell (2,4).
        """

        rc3Cells = array(lrcStroke[:3])  # the 3 cells
        rcMiddle = rc3Cells[1]  # the middle cell which we will update
        bDeadend = np.all(lrcStroke[0] == lrcStroke[2])  # deadend means cell 0 == cell 2

        # get the 2 row, col deltas between the 3 cells, eg [[-1,0],[0,1]] = North, East
        rc2Trans = np.diff(rc3Cells, axis=0)

        # get the direction index for the 2 transitions
        liTrans = []
        for rcTrans in rc2Trans:
            # gRCTrans - rcTrans gives an array of vector differences between our rcTrans
            # and the 4 directions stored in gRCTrans.
            # Where the vector difference is zero, we have a match...
            # np.all detects where the whole row,col vector is zero.
            # argwhere gives the index of the zero vector, ie the direction index
            iTrans = np.argwhere(np.all(self.gRCTrans - rcTrans == 0, axis=1))
            if len(iTrans) > 0:
                iTrans = iTrans[0][0]
                liTrans.append(iTrans)

        # check that we have two transitions
        if len(liTrans) == 2:
            # Set the transition
            # If this transition spans 3 cells, it is not a deadend, so remove any deadends.
            # The user will need to resolve any conflicts.
            self.env.rail.set_transition((*rcMiddle, liTrans[0]),
                                         liTrans[1],
                                         bAddRemove,
                                         remove_deadends=not bDeadend)

            # Also set the reverse transition
            # use the reversed outbound transition for inbound
            # and the reversed inbound transition for outbound
            self.env.rail.set_transition((*rcMiddle, mirror(liTrans[1])),
                                         mirror(liTrans[0]), bAddRemove, remove_deadends=not bDeadend)

        if bPop:
            lrcStroke.pop(0)  # remove the first cell in the stroke

    def mod_rail_2cells(self, lrcCells, bAddRemove=True, iCellToMod=0, bPop=False):
        """
        Add transitions for rail between two cells
        lrcCells -- list of two rc cells
        bAddRemove -- whether to add (True) or remove (False) the transition
        iCellToMod -- the index of the cell to modify: either 0 or 1
        """
        rc2Cells = array(lrcCells[:2])  # the 2 cells
        rcMod = rc2Cells[iCellToMod]  # the cell which we will update

        # get the row, col delta between the 2 cells, eg [-1,0] = North
        rc1Trans = np.diff(rc2Cells, axis=0)

        # get the direction index for the transition
        liTrans = []
        for rcTrans in rc1Trans:
            iTrans = np.argwhere(np.all(self.gRCTrans - rcTrans == 0, axis=1))
            if len(iTrans) > 0:
                iTrans = iTrans[0][0]
                liTrans.append(iTrans)

        # check that we have one transition
        if len(liTrans) == 1:
            # Set the transition as a deadend
            # The transition is going from cell 0 to cell 1.
            if iCellToMod == 0:
                # if 0, reverse the transition, we need to be entering cell 0
                self.env.rail.set_transition((*rcMod, mirror(liTrans[0])), liTrans[0], bAddRemove)
            else:
                # if 1, the transition is entering cell 1
                self.env.rail.set_transition((*rcMod, liTrans[0]), mirror(liTrans[0]), bAddRemove)

        if bPop:
            lrcCells.pop(0)

    def redraw(self):
        self.view.redraw()

    def clear(self):
        self.env.rail.grid[:, :] = 0
        self.env.agents = []
        self.env.agents_static = []

        self.redraw()

    def clear_cell(self, cell_row_col):
        self.debug_cell(cell_row_col)
        self.env.rail.grid[cell_row_col[0], cell_row_col[1]] = 0
        self.redraw()

    def reset(self, replace_agents=False, nAgents=0):
        self.regenerate("complex", nAgents=nAgents)
        self.redraw()

    def restart_agents(self):
        self.env.agents = EnvAgent.list_from_static(self.env.agents_static)
        self.redraw()

    def set_filename(self, filename):
        self.env_filename = filename

    def load(self):
        if os.path.exists(self.env_filename):
            self.log("load file: ", self.env_filename)
            self.env.load(self.env_filename)
            if not self.regen_size_height == self.env.height or not self.regen_size_width == self.env.width:
                self.regen_size_height = self.env.height
                self.regen_size_width = self.env.width
                self.regenerate(None, 0, self.env)
                self.env.load(self.env_filename)

            self.env.restart_agents()
            self.env.reset(False, False)
            self.init_agents_static = None
            self.view.oRT.update_background()
            self.fix_env()
            self.set_env(self.env)
            self.redraw()
        else:
            self.log("File does not exist:", self.env_filename, " Working directory: ", os.getcwd())

    def save(self):
        self.log("save to ", self.env_filename, " working dir: ", os.getcwd())
        temp_store = self.env.agents
        # clear agents before save , because we want the "init" position of the agent to expert
        self.env.agents = []
        self.env.save(self.env_filename)
        # reset agents current (current position)
        self.env.agents = temp_store

    def save_image(self):
        self.view.oRT.gl.save_image('frame_{:04d}.bmp'.format(self.save_image_count))
        self.save_image_count += 1
        self.view.redraw()

    def regenerate(self, method=None, nAgents=0, env=None):
        self.log("Regenerate size",
                 self.regen_size_width,
                 self.regen_size_height)

        if method is None or method == "Empty":
            fnMethod = empty_rail_generator()
        elif method == "Random Cell":
            fnMethod = random_rail_generator(cell_type_relative_proportion=[1] * 11)
        else:
            fnMethod = complex_rail_generator(nr_start_goal=nAgents, nr_extra=20, min_dist=12, seed=int(time.time()))

        if env is None:
            self.env = RailEnv(width=self.regen_size_width,
                               height=self.regen_size_height,
                               rail_generator=fnMethod,
                               number_of_agents=nAgents,
                               obs_builder_object=TreeObsForRailEnv(max_depth=2))
        else:
            self.env = env
        self.env.reset(regen_rail=True)
        self.fix_env()
        self.set_env(self.env)
        self.view.new_env()
        self.redraw()

    def set_regen_width(self, size):
        self.regen_size_width = size

    def set_regen_height(self, size):
        self.regen_size_height = size

    def find_agent_at(self, cell_row_col):
        for agent_idx, agent in enumerate(self.env.agents_static):
            if tuple(agent.position) == tuple(cell_row_col):
                return agent_idx
        return None

    def click_agent(self, cell_row_col):
        """ The user has clicked on a cell -
            - If there is an agent, select it
                - If that agent was already selected, then deselect it
            - If there is no agent selected, and no agent in the cell, create one
            - If there is an agent selected, and no agent in the cell, move the selected agent to the cell
        """

        # Has the user clicked on an existing agent?
        agent_idx = self.find_agent_at(cell_row_col)

        if agent_idx is None:
            # No
            if self.selected_agent is None:
                # Create a new agent and select it.
                agent_static = EnvAgentStatic(position=cell_row_col, direction=0, target=cell_row_col, moving=False)
                self.selected_agent = self.env.add_agent_static(agent_static)
                self.view.oRT.update_background()
            else:
                # Move the selected agent to this cell
                agent_static = self.env.agents_static[self.selected_agent]
                agent_static.position = cell_row_col
                agent_static.old_position = cell_row_col
                self.env.agents = []
        else:
            # Yes
            # Have they clicked on the agent already selected?
            if self.selected_agent is not None and agent_idx == self.selected_agent:
                # Yes - deselect the agent
                self.selected_agent = None
            else:
                # No - select the agent
                self.selected_agent = agent_idx

        self.init_agents_static = None
        self.redraw()

    def add_target(self, rcCell):
        if self.selected_agent is not None:
            self.env.agents_static[self.selected_agent].target = rcCell
            self.init_agents_static = None
            self.view.oRT.update_background()
            self.redraw()

    def fix_env(self):
        self.env.width = self.env.rail.width
        self.env.height = self.env.rail.height

    def log(self, *args, **kwargs):
        if self.view is None:
            print(*args, **kwargs)
        else:
            self.view.log(*args, **kwargs)

    def debug(self, *args, **kwargs):
        if self.debug_bool:
            self.log(*args, **kwargs)

    def debug_cell(self, rcCell):
        binTrans = self.env.rail.get_full_transitions(*rcCell)
        sbinTrans = format(binTrans, "#018b")[2:]
        self.debug("cell ",
                   rcCell,
                   "Transitions: ",
                   binTrans,
                   sbinTrans,
                   [sbinTrans[i:(i + 4)] for i in range(0, len(sbinTrans), 4)])



import io
import os
import platform
import time
import tkinter as tk

import numpy as np
from PIL import Image, ImageDraw, ImageTk  # , ImageFont
from numpy import array
from pkg_resources import resource_string as resource_bytes

from flatland.utils.graphics_layer import GraphicsLayer


def enable_windows_cairo_support():
    if os.name == 'nt':
        import site
        import ctypes.util
        default_os_path = os.environ['PATH']
        os.environ['PATH'] = ''
        for s in site.getsitepackages():
            os.environ['PATH'] = os.environ['PATH'] + ';' + s + '\\cairo'
        os.environ['PATH'] = os.environ['PATH'] + ';' + default_os_path
        if ctypes.util.find_library('cairo') is None:
            print("Error: cairo not installed")


enable_windows_cairo_support()
from cairosvg import svg2png  # noqa: E402
from screeninfo import get_monitors  # noqa: E402

from flatland.core.grid.rail_env_grid import RailEnvTransitions  # noqa: E402


class PILGL(GraphicsLayer):
    # tk.Tk() must be a singleton!
    # https://stackoverflow.com/questions/26097811/image-pyimage2-doesnt-exist
    window = tk.Tk()

    RAIL_LAYER = 0
    PREDICTION_PATH_LAYER = 1
    TARGET_LAYER = 2
    AGENT_LAYER = 3
    SELECTED_AGENT_LAYER = 4
    SELECTED_TARGET_LAYER = 5

    def __init__(self, width, height, jupyter=False):
        self.yxBase = (0, 0)
        self.linewidth = 4
        self.n_agent_colors = 1  # overridden in loadAgent

        self.width = width
        self.height = height

        self.background_grid = np.zeros(shape=(self.width, self.height))

        if jupyter is False:
            self.screen_width = 800
            self.screen_height = 600

            if platform.system() == "Windows" or platform.system() == "Linux":
                self.screen_width = 9999
                self.screen_height = 9999
                for m in get_monitors():
                    self.screen_height = min(self.screen_height, m.height)
                    self.screen_width = min(self.screen_width, m.width)
                # Note: screeninfo doesnot have proper support for 
                # OSX yet, hence the default values of 800,600
                # will be used for the same.

            w = (self.screen_width - self.width - 10) / (self.width + 1 + self.linewidth)
            h = (self.screen_height - self.height - 10) / (self.height + 1 + self.linewidth)
            self.nPixCell = int(max(1, np.ceil(min(w, h))))
        else:
            self.nPixCell = 40

        # Total grid size at native scale
        self.widthPx = self.width * self.nPixCell + self.linewidth
        self.heightPx = self.height * self.nPixCell + self.linewidth

        self.xPx = int((self.screen_width - self.widthPx) / 2.0)
        self.yPx = int((self.screen_height - self.heightPx) / 2.0)

        self.layers = []
        self.draws = []

        self.tColBg = (255, 255, 255)  # white background
        self.tColRail = (0, 0, 0)  # black rails
        self.tColGrid = (230,) * 3  # light grey for grid

        sColors = "d50000#c51162#aa00ff#6200ea#304ffe#2962ff#0091ea#00b8d4#00bfa5#00c853" + \
                  "#64dd17#aeea00#ffd600#ffab00#ff6d00#ff3d00#5d4037#455a64"

        self.agent_colors = [self.rgb_s2i(sColor) for sColor in sColors.split("#")]
        self.n_agent_colors = len(self.agent_colors)

        self.window_open = False
        self.firstFrame = True
        self.old_background_image = (None, None, None)
        self.create_layers()

    def build_background_map(self, dTargets):
        x = self.old_background_image
        rebuild = False
        if x[0] is None:
            rebuild = True
        else:
            if len(x[0]) != len(dTargets):
                rebuild = True
            else:
                if x[0] != dTargets:
                    rebuild = True
                if x[1] != self.width:
                    rebuild = True
                if x[2] != self.height:
                    rebuild = True

        if rebuild:
            self.background_grid = np.zeros(shape=(self.width, self.height))
            for x in range(self.width):
                for y in range(self.height):
                    distance = int(np.ceil(np.sqrt(self.width ** 2.0 + self.height ** 2.0)))
                    for rc in dTargets:
                        r = rc[1]
                        c = rc[0]
                        d = int(np.floor(np.sqrt((x - r) ** 2 + (y - c) ** 2)))
                        distance = min(d, distance)
                    self.background_grid[x][y] = distance

            self.old_background_image = (dTargets, self.width, self.height)

    def rgb_s2i(self, sRGB):
        """ convert a hex RGB string like 0091ea to 3-tuple of ints """
        return tuple(int(sRGB[iRGB * 2:iRGB * 2 + 2], 16) for iRGB in [0, 1, 2])

    def get_agent_color(self, iAgent):
        return self.agent_colors[iAgent % self.n_agent_colors]

    def plot(self, gX, gY, color=None, linewidth=3, layer=RAIL_LAYER, opacity=255, **kwargs):
        color = self.adapt_color(color)
        if len(color) == 3:
            color += (opacity,)
        elif len(color) == 4:
            color = color[:3] + (opacity,)
        gPoints = np.stack([array(gX), -array(gY)]).T * self.nPixCell
        gPoints = list(gPoints.ravel())
        self.draws[layer].line(gPoints, fill=color, width=self.linewidth)

    def scatter(self, gX, gY, color=None, marker="o", s=50, layer=RAIL_LAYER, opacity=255, *args, **kwargs):
        color = self.adapt_color(color)
        r = np.sqrt(s)
        gPoints = np.stack([np.atleast_1d(gX), -np.atleast_1d(gY)]).T * self.nPixCell
        for x, y in gPoints:
            self.draws[layer].rectangle([(x - r, y - r), (x + r, y + r)], fill=color, outline=color)

    def draw_image_xy(self, pil_img, xyPixLeftTop, layer=RAIL_LAYER, ):
        if (pil_img.mode == "RGBA"):
            pil_mask = pil_img
        else:
            pil_mask = None

        self.layers[layer].paste(pil_img, xyPixLeftTop, pil_mask)

    def draw_image_row_col(self, pil_img, rcTopLeft, layer=RAIL_LAYER, ):
        xyPixLeftTop = tuple((array(rcTopLeft) * self.nPixCell)[[1, 0]])
        self.draw_image_xy(pil_img, xyPixLeftTop, layer=layer)

    def open_window(self):
        assert self.window_open is False, "Window is already open!"
        self.__class__.window.title("Flatland")
        self.__class__.window.configure(background='grey')
        self.window_open = True

    def close_window(self):
        self.panel.destroy()
        # quit but not destroy!
        self.__class__.window.quit()

    def text(self, *args, **kwargs):
        pass

    def prettify(self, *args, **kwargs):
        pass

    def prettify2(self, width, height, cell_size):
        pass

    def begin_frame(self):
        # Create a new agent layer
        self.create_layer(iLayer=PILGL.AGENT_LAYER, clear=True)
        self.create_layer(iLayer=PILGL.PREDICTION_PATH_LAYER, clear=True)

    def show(self, block=False):
        img = self.alpha_composite_layers()

        if not self.window_open:
            self.open_window()

        tkimg = ImageTk.PhotoImage(img)

        if self.firstFrame:
            # Do TK actions for a new panel (not sure what they really do)
            self.panel = tk.Label(self.window, image=tkimg)
            self.panel.pack(side="bottom", fill="both", expand="yes")
        else:
            # update the image in situ
            self.panel.configure(image=tkimg)
            self.panel.image = tkimg

        self.__class__.window.update()
        self.firstFrame = False

    def pause(self, seconds=0.00001):
        pass

    def alpha_composite_layers(self):
        img = self.layers[0]
        for img2 in self.layers[1:]:
            img = Image.alpha_composite(img, img2)
        return img

    def get_image(self):
        """ return a blended / alpha composited image composed of all the layers,
            with layer 0 at the "back".
        """
        img = self.alpha_composite_layers()
        return array(img)

    def save_image(self, filename):
        """
        Renders the current scene into a image file
        :param filename: filename where to store the rendering output_generator
        (supported image format *.bmp , .. , *.png)
        """
        img = self.alpha_composite_layers()
        img.save(filename)

    def create_image(self, opacity=255):
        img = Image.new("RGBA", (self.widthPx, self.heightPx), (255, 255, 255, opacity))
        return img

    def clear_layer(self, iLayer=0, opacity=None):
        if opacity is None:
            opacity = 0 if iLayer > 0 else 255
        self.layers[iLayer] = img = self.create_image(opacity)
        # We also need to maintain a Draw object for each layer
        self.draws[iLayer] = ImageDraw.Draw(img)

    def create_layer(self, iLayer=0, clear=True):
        # If we don't have the layers already, create them
        if len(self.layers) <= iLayer:
            for i in range(len(self.layers), iLayer + 1):
                if i == 0:
                    opacity = 255  # "bottom" layer is opaque (for rails)
                else:
                    opacity = 0  # subsequent layers are transparent
                img = self.create_image(opacity)
                self.layers.append(img)
                self.draws.append(ImageDraw.Draw(img))
        else:
            # We do already have this iLayer.  Clear it if requested.
            if clear:
                self.clear_layer(iLayer)

    def create_layers(self, clear=True):
        self.create_layer(PILGL.RAIL_LAYER, clear=clear)  # rail / background (scene)
        self.create_layer(PILGL.AGENT_LAYER, clear=clear)  # agents
        self.create_layer(PILGL.TARGET_LAYER, clear=clear)  # agents
        self.create_layer(PILGL.PREDICTION_PATH_LAYER, clear=clear)  # drawing layer for agent's prediction path
        self.create_layer(PILGL.SELECTED_AGENT_LAYER, clear=clear)  # drawing layer for selected agent
        self.create_layer(PILGL.SELECTED_TARGET_LAYER, clear=clear)  # drawing layer for selected agent's target


class PILSVG(PILGL):
    def __init__(self, width, height, jupyter=False):
        oSuper = super()
        oSuper.__init__(width, height, jupyter)

        self.lwAgents = []
        self.agents_prev = []

        self.load_buildings()
        self.load_scenery()
        self.load_rail()
        self.load_agent()

    def process_events(self):
        time.sleep(0.001)

    def clear_rails(self):
        self.create_layers()
        self.clear_agents()

    def clear_agents(self):
        for wAgent in self.lwAgents:
            self.layout.removeWidget(wAgent)
        self.lwAgents = []
        self.agents_prev = []

    def pil_from_svg_file(self, package, resource):
        bytestring = resource_bytes(package, resource)
        bytesPNG = svg2png(bytestring=bytestring, output_height=self.nPixCell, output_width=self.nPixCell)
        with io.BytesIO(bytesPNG) as fIn:
            pil_img = Image.open(fIn)
            pil_img.load()

        return pil_img

    def pil_from_svg_bytes(self, bytesSVG):
        bytesPNG = svg2png(bytesSVG, output_height=self.nPixCell, output_width=self.nPixCell)
        with io.BytesIO(bytesPNG) as fIn:
            pil_img = Image.open(fIn)
            return pil_img

    def load_buildings(self):
        dBuildingFiles = [
            "Buildings-Bank.svg",
            "Buildings-Bar.svg",
            "Buildings-Wohnhaus.svg",
            "Buildings-Hochhaus.svg",
            "Buildings-Hotel.svg",
            "Buildings-Office.svg",
            "Buildings-Polizei.svg",
            "Buildings-Post.svg",
            "Buildings-Supermarkt.svg",
            "Buildings-Tankstelle.svg",
            "Buildings-Fabrik_A.svg",
            "Buildings-Fabrik_B.svg",
            "Buildings-Fabrik_C.svg",
            "Buildings-Fabrik_D.svg",
            "Buildings-Fabrik_E.svg",
            "Buildings-Fabrik_F.svg",
            "Buildings-Fabrik_G.svg",
            "Buildings-Fabrik_H.svg",
            "Buildings-Fabrik_I.svg"
        ]

        imgBg = self.pil_from_svg_file('svg', "Background_city.svg")

        self.dBuildings = []
        for sFile in dBuildingFiles:
            img = self.pil_from_svg_file('svg', sFile)
            img = Image.alpha_composite(imgBg, img)
            self.dBuildings.append(img)

    def load_scenery(self):
        scenery_files = [
            "Scenery-Laubbaume_A.svg",
            "Scenery-Laubbaume_B.svg",
            "Scenery-Laubbaume_C.svg",
            "Scenery-Nadelbaume_A.svg",
            "Scenery-Nadelbaume_B.svg",
            "Scenery-Bergwelt_B.svg"
        ]

        scenery_files_d2 = [
            "Scenery-Bergwelt_C_Teil_1_links.svg",
            "Scenery-Bergwelt_C_Teil_2_rechts.svg"
        ]

        scenery_files_d3 = [
            "Scenery-Bergwelt_A_Teil_3_rechts.svg",
            "Scenery-Bergwelt_A_Teil_2_mitte.svg",
            "Scenery-Bergwelt_A_Teil_1_links.svg"
        ]

        img_back_ground = self.pil_from_svg_file('svg', "Background_Light_green.svg")

        self.scenery = []
        for file in scenery_files:
            img = self.pil_from_svg_file('svg', file)
            img = Image.alpha_composite(img_back_ground, img)
            self.scenery.append(img)

        self.scenery_d2 = []
        for file in scenery_files_d2:
            img = self.pil_from_svg_file('svg', file)
            img = Image.alpha_composite(img_back_ground, img)
            self.scenery_d2.append(img)

        self.scenery_d3 = []
        for file in scenery_files_d3:
            img = self.pil_from_svg_file('svg', file)
            img = Image.alpha_composite(img_back_ground, img)
            self.scenery_d3.append(img)

    def load_rail(self):
        """ Load the rail SVG images, apply rotations, and store as PIL images.
        """
        rail_files = {
            "": "Background_Light_green.svg",
            "WE": "Gleis_Deadend.svg",
            "WW EE NN SS": "Gleis_Diamond_Crossing.svg",
            "WW EE": "Gleis_horizontal.svg",
            "EN SW": "Gleis_Kurve_oben_links.svg",
            "WN SE": "Gleis_Kurve_oben_rechts.svg",
            "ES NW": "Gleis_Kurve_unten_links.svg",
            "NE WS": "Gleis_Kurve_unten_rechts.svg",
            "NN SS": "Gleis_vertikal.svg",
            "NN SS EE WW ES NW SE WN": "Weiche_Double_Slip.svg",
            "EE WW EN SW": "Weiche_horizontal_oben_links.svg",
            "EE WW SE WN": "Weiche_horizontal_oben_rechts.svg",
            "EE WW ES NW": "Weiche_horizontal_unten_links.svg",
            "EE WW NE WS": "Weiche_horizontal_unten_rechts.svg",
            "NN SS EE WW NW ES": "Weiche_Single_Slip.svg",
            "NE NW ES WS": "Weiche_Symetrical.svg",
            "NN SS EN SW": "Weiche_vertikal_oben_links.svg",
            "NN SS SE WN": "Weiche_vertikal_oben_rechts.svg",
            "NN SS NW ES": "Weiche_vertikal_unten_links.svg",
            "NN SS NE WS": "Weiche_vertikal_unten_rechts.svg",
            "NE NW ES WS SS NN": "Weiche_Symetrical_gerade.svg",
            "NE EN SW WS": "Gleis_Kurve_oben_links_unten_rechts.svg"
        }

        target_files = {
            "EW": "Bahnhof_#d50000_Deadend_links.svg",
            "NS": "Bahnhof_#d50000_Deadend_oben.svg",
            "WE": "Bahnhof_#d50000_Deadend_rechts.svg",
            "SN": "Bahnhof_#d50000_Deadend_unten.svg",
            "EE WW": "Bahnhof_#d50000_Gleis_horizontal.svg",
            "NN SS": "Bahnhof_#d50000_Gleis_vertikal.svg"}

        # Dict of rail cell images indexed by binary transitions
        pil_rail_files_org = self.load_svgs(rail_files, rotate=True)
        pil_rail_files = self.load_svgs(rail_files, rotate=True, background_image="Background_rail.svg",
                                        whitefilter="Background_white_filter.svg")

        # Load the target files (which have rails and transitions of their own)
        # They are indexed by (binTrans, iAgent), ie a tuple of the binary transition and the agent index
        pil_target_files_org = self.load_svgs(target_files, rotate=False, agent_colors=self.agent_colors)
        pil_target_files = self.load_svgs(target_files, rotate=False, agent_colors=self.agent_colors,
                                          background_image="Background_rail.svg",
                                          whitefilter="Background_white_filter.svg")

        # Load station and recolorize them
        station = self.pil_from_svg_file("svg", "Bahnhof_#d50000_target.svg")
        self.station_colors = self.recolor_image(station, [0, 0, 0], self.agent_colors, False)

        cell_occupied = self.pil_from_svg_file("svg", "Cell_occupied.svg")
        self.cell_occupied = self.recolor_image(cell_occupied, [0, 0, 0], self.agent_colors, False)

        # Merge them with the regular rails.
        # https://stackoverflow.com/questions/38987/how-to-merge-two-dictionaries-in-a-single-expression
        self.pil_rail = {**pil_rail_files, **pil_target_files}
        self.pil_rail_org = {**pil_rail_files_org, **pil_target_files_org}

    def load_svgs(self, file_directory, rotate=False, agent_colors=False, background_image=None, whitefilter=None):
        pil = {}

        transitions = RailEnvTransitions()

        directions = list("NESW")

        for transition, file in file_directory.items():

            # Translate the ascii transition description in the format  "NE WS" to the 
            # binary list of transitions as per RailEnv - NESW (in) x NESW (out)
            transition_16_bit = ["0"] * 16
            for sTran in transition.split(" "):
                if len(sTran) == 2:
                    in_direction = directions.index(sTran[0])
                    out_direction = directions.index(sTran[1])
                    transition_idx = 4 * in_direction + out_direction
                    transition_16_bit[transition_idx] = "1"
            transition_16_bit_string = "".join(transition_16_bit)
            binary_trans = int(transition_16_bit_string, 2)

            pil_rail = self.pil_from_svg_file('svg', file)

            if background_image is not None:
                img_bg = self.pil_from_svg_file('svg', background_image)
                pil_rail = Image.alpha_composite(img_bg, pil_rail)

            if whitefilter is not None:
                img_bg = self.pil_from_svg_file('svg', whitefilter)
                pil_rail = Image.alpha_composite(pil_rail, img_bg)

            if rotate:
                # For rotations, we also store the base image
                pil[binary_trans] = pil_rail
                # Rotate both the transition binary and the image and save in the dict
                for nRot in [90, 180, 270]:
                    binary_trans_2 = transitions.rotate_transition(binary_trans, nRot)

                    # PIL rotates anticlockwise for positive theta
                    pil_rail_2 = pil_rail.rotate(-nRot)
                    pil[binary_trans_2] = pil_rail_2

            if agent_colors:
                # For recoloring, we don't store the base image.
                base_color = self.rgb_s2i("d50000")
                pils = self.recolor_image(pil_rail, base_color, self.agent_colors)
                for color_idx, pil_rail_2 in enumerate(pils):
                    pil[(binary_trans, color_idx)] = pils[color_idx]

        return pil

    def set_predicion_path_at(self, row, col, binary_trans, agent_rail_color):
        colored_rail = self.recolor_image(self.pil_rail_org[binary_trans],
                                          [61, 61, 61], [agent_rail_color],
                                          False)[0]
        self.draw_image_row_col(colored_rail, (row, col), layer=PILGL.PREDICTION_PATH_LAYER)

    def set_rail_at(self, row, col, binary_trans, target=None, is_selected=False, rail_grid=None):
        if binary_trans in self.pil_rail:
            pil_track = self.pil_rail[binary_trans]
            if target is not None:
                target_img = self.station_colors[target % len(self.station_colors)]
                target_img = Image.alpha_composite(pil_track, target_img)
                self.draw_image_row_col(target_img, (row, col), layer=PILGL.TARGET_LAYER)

            if binary_trans == 0:
                if self.background_grid[col][row] <= 4:
                    a = int(self.background_grid[col][row])
                    a = a % len(self.dBuildings)
                    if (col + row + col * row) % 13 > 11:
                        pil_track = self.scenery[a % len(self.scenery)]
                    else:
                        if (col + row + col * row) % 3 == 0:
                            a = (a + (col + row + col * row)) % len(self.dBuildings)
                        pil_track = self.dBuildings[a]
                elif (self.background_grid[col][row] > 4) or ((col ** 3 + row ** 2 + col * row) % 10 == 0):
                    a = int(self.background_grid[col][row]) - 4
                    a2 = (a + (col + row + col * row + col ** 3 + row ** 4))
                    if a2 % 17 > 11:
                        a = a2
                    pil_track = self.scenery[a % len(self.scenery)]

            self.draw_image_row_col(pil_track, (row, col), layer=PILGL.RAIL_LAYER)
        else:
            print("Illegal rail:", row, col, format(binary_trans, "#018b")[2:], binary_trans)

        if target is not None:
            if is_selected:
                svgBG = self.pil_from_svg_file("svg", "Selected_Target.svg")
                self.clear_layer(PILGL.SELECTED_TARGET_LAYER, 0)
                self.draw_image_row_col(svgBG, (row, col), layer=PILGL.SELECTED_TARGET_LAYER)

    def recolor_image(self, pil, a3BaseColor, ltColors, invert=False):
        rgbaImg = array(pil)
        pils = []
        for iColor, tnColor in enumerate(ltColors):
            # find the pixels which match the base paint color
            if invert:
                xy_color_mask = np.all(rgbaImg[:, :, 0:3] - a3BaseColor != 0, axis=2)
            else:
                xy_color_mask = np.all(rgbaImg[:, :, 0:3] - a3BaseColor == 0, axis=2)
            rgbaImg2 = np.copy(rgbaImg)

            # Repaint the base color with the new color
            rgbaImg2[xy_color_mask, 0:3] = tnColor
            pil2 = Image.fromarray(rgbaImg2)
            pils.append(pil2)
        return pils

    def load_agent(self):

        # Seed initial train/zug files indexed by tuple(iDirIn, iDirOut):
        file_directory = {
            (0, 0): "Zug_Gleis_#0091ea.svg",
            (1, 2): "Zug_1_Weiche_#0091ea.svg",
            (0, 3): "Zug_2_Weiche_#0091ea.svg"
        }

        # "paint" color of the train images we load - this is the color we will change.
        # base_color = self.rgb_s2i("0091ea") \#  noqa: E800
        # temporary workaround for trains / agents renamed with different colour:
        base_color = self.rgb_s2i("d50000")

        self.pil_zug = {}

        for directions, path_svg in file_directory.items():
            in_direction, out_direction = directions

            pil_zug = self.pil_from_svg_file("svg", path_svg)

            # Rotate both the directions and the image and save in the dict
            for rot_direction in range(4):
                rotation_degree = rot_direction * 90
                in_direction_2 = (in_direction + rot_direction) % 4
                out_direction_2 = (out_direction + rot_direction) % 4

                # PIL rotates anticlockwise for positive theta
                pil_zug_2 = pil_zug.rotate(-rotation_degree)

                # Save colored versions of each rotation / variant
                pils = self.recolor_image(pil_zug_2, base_color, self.agent_colors)
                for color_idx, pil_zug_3 in enumerate(pils):
                    self.pil_zug[(in_direction_2, out_direction_2, color_idx)] = pils[color_idx]

    def set_agent_at(self, agent_idx, row, col, in_direction, out_direction, is_selected):
        delta_dir = (out_direction - in_direction) % 4
        color_idx = agent_idx % self.n_agent_colors
        # when flipping direction at a dead end, use the "out_direction" direction.
        if delta_dir == 2:
            in_direction = out_direction
        pil_zug = self.pil_zug[(in_direction % 4, out_direction % 4, color_idx)]
        self.draw_image_row_col(pil_zug, (row, col), layer=PILGL.AGENT_LAYER)

        if is_selected:
            bg_svg = self.pil_from_svg_file("svg", "Selected_Agent.svg")
            self.clear_layer(PILGL.SELECTED_AGENT_LAYER, 0)
            self.draw_image_row_col(bg_svg, (row, col), layer=PILGL.SELECTED_AGENT_LAYER)

    def set_cell_occupied(self, agent_idx, row, col):
        occupied_im = self.cell_occupied[agent_idx % len(self.cell_occupied)]
        self.draw_image_row_col(occupied_im, (row, col), 1)


def main2():
    gl = PILSVG(10, 10)
    for i in range(10):
        gl.begin_frame()
        gl.plot([3 + i, 4], [-4 - i, -5], color="r")
        gl.endFrame()
        time.sleep(1)


def main():
    gl = PILSVG(width=10, height=10)

    for i in range(1000):
        gl.process_events()
        time.sleep(0.1)
    time.sleep(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for `flatland` package.
"""

import sys

import numpy as np
from importlib_resources import path

import flatland.utils.rendertools as rt
import images.test
from flatland.envs.generators import empty_rail_generator
from flatland.envs.observations import TreeObsForRailEnv
from flatland.envs.rail_env import RailEnv


def checkFrozenImage(oRT, sFileImage, resave=False):
    sDirRoot = "."
    sDirImages = sDirRoot + "/images/"

    img_test = oRT.get_image()

    if resave:
        np.savez_compressed(sDirImages + sFileImage, img=img_test)
        return

    with path(images, sFileImage) as file_in:
        np.load(file_in)

    # TODO fails!
    #  assert (img_test.shape == img_expected.shape) \ #  noqa: E800
    #  assert ((np.sum(np.square(img_test - img_expected)) / img_expected.size / 256) < 1e-3), \ #  noqa: E800
    #      "Image {} does not match".format(sFileImage) \ #  noqa: E800


def test_render_env(save_new_images=False):
    np.random.seed(100)
    oEnv = RailEnv(width=10, height=10,
                   rail_generator=empty_rail_generator(),
                   number_of_agents=0,
                   obs_builder_object=TreeObsForRailEnv(max_depth=2)
                   )
    oEnv.rail.load_transition_map('env_data.tests', "test1.npy")
    oRT = rt.RenderTool(oEnv, gl="PILSVG")
    oRT.render_env(show=False)

    checkFrozenImage(oRT, "basic-env.npz", resave=save_new_images)

    oRT = rt.RenderTool(oEnv, gl="PIL")
    oRT.render_env()
    checkFrozenImage(oRT, "basic-env-PIL.npz", resave=save_new_images)


def main():
    if len(sys.argv) == 2 and sys.argv[1] == "save":
        test_render_env(save_new_images=True)
    else:
        print("Run 'python test_flatland_utils_rendertools.py save' to regenerate images")
        test_render_env()


if __name__ == "__main__":
    main()



import copy
import re

import svgutils

from flatland.core.grid.rail_env_grid import RailEnvTransitions


class SVG(object):
    def __init__(self, sfName=None, svgETree=None):

        if sfName is not None:
            self.svg = svgutils.transform.fromfile(sfName)
        elif svgETree is not None:
            self.svg = svgETree

        expr = "//*[local-name() = $name]"
        self.eStyle = self.svg.root.xpath(expr, name="style")[0]
        ltMatch = re.findall(r".st([a-zA-Z0-9]+)[{]([^}]*)}", self.eStyle.text)
        self.dStyles = dict(ltMatch)

    def copy(self):
        new_svg = copy.deepcopy(self.svg)

        return SVG(svgETree=new_svg)

    def merge(self, svg2):
        svg3 = svg2.copy()

        svg3.renumber_styles(offset=10)
        svg3.eStyle.text = self.eStyle.text + "\n" + svg3.eStyle.text

        for child in self.svg.root:
            if not child.tag.endswith("style"):
                svg3.svg.root.append(child)

        return svg3

    def renumber_styles(self, offset=10):
        sNewStyles = "\n"
        for sStyle in self.dStyles.keys():
            iStyle = int(sStyle)
            sClass = "st{:}".format(iStyle)
            lEl = self.svg.root.xpath("//*[@class='{}']".format(sClass))
            for el in lEl:
                el.attrib["class"] = "st{}".format(iStyle + offset)

            sStyle2 = str(iStyle + offset)

            sNewStyle = "\t.st" + sStyle2 + "{" + self.dStyles[sStyle] + "}\n"
            sNewStyles += sNewStyle

        self.eStyle.text = sNewStyles

    def set_style_color(self, style_name, color):
        sNewStyles = "\n"
        for sKey, sValue in self.dStyles.items():
            if sKey == style_name:
                sValue = "fill:#" + "".join([('{:#04x}'.format(int(255.0 * col))[2:4]) for col in color[0:3]]) + ";"
            sNewStyle = "\t.st" + sKey + "{" + sValue + "}\n"
            sNewStyles += sNewStyle

        self.eStyle.text = sNewStyles

    def set_rotate(self, angle):
        self.svg.root.attrib["transform"] = "rotate({}, 120, 120)".format(angle)

    def to_string(self):
        return self.svg.to_str().decode("utf-8")


class Zug(object):
    def __init__(self, iDir=0):
        self.svg_straight = SVG("svg/Zug_Gleis_#0091ea.svg")
        self.svg_curve1 = SVG("svg/Zug_1_Weiche_#0091ea.svg")
        self.svg_curve2 = SVG("svg/Zug_2_Weiche_#0091ea.svg")

    def getSvg(self, iAgent, iDirIn, iDirOut, color=None):

        delta_dir = (iDirOut - iDirIn) % 4

        if delta_dir in (0, 2):
            svg = self.svg_straight.copy()
            svg.set_rotate(iDirIn * 90)

        if delta_dir == 1:  # bend to right, eg N->E, E->S
            svg = self.svg_curve1.copy()
            svg.set_rotate((iDirIn - 1) * 90)

        elif delta_dir == 3:  # bend to left, eg N->W
            svg = self.svg_curve2.copy()
            svg.set_rotate(iDirIn * 90)

        if color is not None:
            svg.set_style_color("2", color)

        return svg


class Track(object):
    """ Class to load and hold SVG track images.
        Creates a mapping between
        - cell entry and exit directions (ie transitions), and
        - specific images provided by the SBB graphic artist.
        The directions and images are also rotated by 90, 180 & 270 degrees.
        (There is some redundancy in this process, given the images provided)
    """

    def __init__(self):
        dFiles = {
            "": "Background_#9CCB89.svg",
            "WE": "Gleis_Deadend.svg",
            "WW EE NN SS": "Gleis_Diamond_Crossing.svg",
            "WW EE": "Gleis_horizontal.svg",
            "EN SW": "Gleis_Kurve_oben_links.svg",
            "WN SE": "Gleis_Kurve_oben_rechts.svg",
            "ES NW": "Gleis_Kurve_unten_links.svg",
            "NE WS": "Gleis_Kurve_unten_rechts.svg",
            "NN SS": "Gleis_vertikal.svg",
            "NN SS EE WW ES NW SE WN": "Weiche_Double_Slip.svg",
            "EE WW EN SW": "Weiche_horizontal_oben_links.svg",
            "EE WW SE WN": "Weiche_horizontal_oben_rechts.svg",
            "EE WW ES NW": "Weiche_horizontal_unten_links.svg",
            "EE WW NE WS": "Weiche_horizontal_unten_rechts.svg",
            "NN SS EE WW NW ES": "Weiche_Single_Slip.svg",
            "NE NW ES WS": "Weiche_Symetrical.svg",
            "NN SS EN SW": "Weiche_vertikal_oben_links.svg",
            "NN SS SE WN": "Weiche_vertikal_oben_rechts.svg",
            "NN SS NW ES": "Weiche_vertikal_unten_links.svg",
            "NN SS NE WS": "Weiche_vertikal_unten_rechts.svg"}

        self.dSvg = {}

        transitions = RailEnvTransitions()

        lDirs = list("NESW")

        svgBG = SVG("./svg/Background_#9CCB89.svg")

        for sTrans, sFile in dFiles.items():
            svg = SVG("./svg/" + sFile)

            # Translate the ascii transition descption in the format  "NE WS" to the 
            # binary list of transitions as per RailEnv - NESW (in) x NESW (out)
            lTrans16 = ["0"] * 16
            for sTran in sTrans.split(" "):
                if len(sTran) == 2:
                    iDirIn = lDirs.index(sTran[0])
                    iDirOut = lDirs.index(sTran[1])
                    iTrans = 4 * iDirIn + iDirOut
                    lTrans16[iTrans] = "1"
            sTrans16 = "".join(lTrans16)
            binTrans = int(sTrans16, 2)
            print(sTrans, sTrans16, sFile)

            # Merge the transition svg image with the background colour.
            # This is a shortcut / hack and will need re-working.
            if binTrans > 0:
                svg = svg.merge(svgBG)

            self.dSvg[binTrans] = svg

            # Rotate both the transition binary and the image and save in the dict
            for nRot in [90, 180, 270]:
                binTrans2 = transitions.rotate_transition(binTrans, nRot)
                svg2 = svg.copy()
                svg2.set_rotate(nRot)
                self.dSvg[binTrans2] = svg2


def main():
    zug = Zug()

    svg = zug.getSvg(0, 0, 0, color=(255, 0, 0))
    print(svg.to_string()[:800])


if __name__ == "__main__":
    main()



import random

import numpy as np

from flatland.envs.generators import random_rail_generator
from flatland.envs.rail_env import RailEnv
from flatland.utils.rendertools import RenderTool

random.seed(100)
np.random.seed(100)

# Relative weights of each cell type to be used by the random rail generators.
transition_probability = [1.0,  # empty cell - Case 0
                          1.0,  # Case 1 - straight
                          1.0,  # Case 2 - simple switch
                          0.3,  # Case 3 - diamond drossing
                          0.5,  # Case 4 - single slip
                          0.5,  # Case 5 - double slip
                          0.2,  # Case 6 - symmetrical
                          0.0,  # Case 7 - dead end
                          0.2,  # Case 8 - turn left
                          0.2,  # Case 9 - turn right
                          1.0]  # Case 10 - mirrored switch

# Example generate a random rail
env = RailEnv(width=10,
              height=10,
              rail_generator=random_rail_generator(cell_type_relative_proportion=transition_probability),
              number_of_agents=3)

env.reset()

env_renderer = RenderTool(env, gl="PIL")
env_renderer.render_env(show=True)

input("Press Enter to continue...")



from flatland.core.grid.grid4 import Grid4TransitionsEnum


def get_direction(pos1, pos2) -> Grid4TransitionsEnum:
    """
    Assumes pos1 and pos2 are adjacent location on grid.
    Returns direction (int) that can be used with transitions.
    """
    diff_0 = pos2[0] - pos1[0]
    diff_1 = pos2[1] - pos1[1]
    if diff_0 < 0:
        return 0
    if diff_0 > 0:
        return 2
    if diff_1 > 0:
        return 1
    if diff_1 < 0:
        return 3
    raise Exception("Could not determine direction {}->{}".format(pos1, pos2))


def mirror(dir):
    return (dir + 2) % 4


def validate_new_transition(rail_trans, rail_array, prev_pos, current_pos, new_pos, end_pos):
    # start by getting direction used to get to current node
    # and direction from current node to possible child node
    new_dir = get_direction(current_pos, new_pos)
    if prev_pos is not None:
        current_dir = get_direction(prev_pos, current_pos)
    else:
        current_dir = new_dir
    # create new transition that would go to child
    new_trans = rail_array[current_pos]
    if prev_pos is None:
        if new_trans == 0:
            # need to flip direction because of how end points are defined
            new_trans = rail_trans.set_transition(new_trans, mirror(current_dir), new_dir, 1)
        else:
            # check if matches existing layout
            new_trans = rail_trans.set_transition(new_trans, current_dir, new_dir, 1)
    else:
        # set the forward path
        new_trans = rail_trans.set_transition(new_trans, current_dir, new_dir, 1)
        # set the backwards path
        new_trans = rail_trans.set_transition(new_trans, mirror(new_dir), mirror(current_dir), 1)
    if new_pos == end_pos:
        # need to validate end pos setup as well
        new_trans_e = rail_array[end_pos]
        if new_trans_e == 0:
            # need to flip direction because of how end points are defined
            new_trans_e = rail_trans.set_transition(new_trans_e, new_dir, mirror(new_dir), 1)
        else:
            # check if matches existing layout
            new_trans_e = rail_trans.set_transition(new_trans_e, new_dir, new_dir, 1)

        if not rail_trans.is_valid(new_trans_e):
            return False

    # is transition is valid?
    return rail_trans.is_valid(new_trans)


def get_new_position(position, movement):
    """ Utility function that converts a compass movement over a 2D grid to new positions (r, c). """
    if movement == Grid4TransitionsEnum.NORTH:
        return (position[0] - 1, position[1])
    elif movement == Grid4TransitionsEnum.EAST:
        return (position[0], position[1] + 1)
    elif movement == Grid4TransitionsEnum.SOUTH:
        return (position[0] + 1, position[1])
    elif movement == Grid4TransitionsEnum.WEST:
        return (position[0], position[1] - 1)



import redis
import json
import os
import numpy as np
import msgpack
import msgpack_numpy as m
import hashlib
import random
from flatland.evaluators import messages
from flatland.envs.rail_env import RailEnv
from flatland.envs.generators import rail_from_file
from flatland.envs.observations import TreeObsForRailEnv
from flatland.envs.predictions import ShortestPathPredictorForRailEnv
import time
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
m.patch()


def are_dicts_equal(d1, d2):
    """ return True if all keys and values are the same """
    return all(k in d2 and d1[k] == d2[k]
               for k in d1) \
        and all(k in d1 and d1[k] == d2[k]
               for k in d2)


class FlatlandRemoteClient(object):
    """
        Redis client to interface with flatland-rl remote-evaluation-service
        The Docker container hosts a redis-server inside the container.
        This client connects to the same redis-server, 
        and communicates with the service.
        The service eventually will reside outside the docker container, 
        and will communicate
        with the client only via the redis-server of the docker container.
        On the instantiation of the docker container, one service will be 
        instantiated parallely.
        The service will accepts commands at "`service_id`::commands"
        where `service_id` is either provided as an `env` variable or is
        instantiated to "flatland_rl_redis_service_id"
    """
    def __init__(self,  
                remote_host='127.0.0.1',
                remote_port=6379,
                remote_db=0,
                remote_password=None,
                test_envs_root=None,
                verbose=False):

        self.remote_host = remote_host
        self.remote_port = remote_port
        self.remote_db = remote_db
        self.remote_password = remote_password
        self.redis_pool = redis.ConnectionPool(
                                host=remote_host,
                                port=remote_port,
                                db=remote_db,
                                password=remote_password)
        self.namespace = "flatland-rl"
        self.service_id = os.getenv(
                            'FLATLAND_RL_SERVICE_ID',
                            'FLATLAND_RL_SERVICE_ID'
                            )
        self.command_channel = "{}::{}::commands".format(
                                    self.namespace,
                                    self.service_id
                                )
        if test_envs_root:
            self.test_envs_root = test_envs_root
        else:
            self.test_envs_root = os.getenv(
                                'AICROWD_TESTS_FOLDER',
                                '/tmp/flatland_envs'
                                )

        self.verbose = verbose

        self.env = None
        self.ping_pong()

    def get_redis_connection(self):
        return redis.Redis(connection_pool=self.redis_pool)

    def _generate_response_channel(self):
        random_hash = hashlib.md5(
                        "{}".format(
                                random.randint(0, 10**10)
                            ).encode('utf-8')).hexdigest()
        response_channel = "{}::{}::response::{}".format(self.namespace,
                                                        self.service_id,
                                                        random_hash)
        return response_channel

    def _blocking_request(self, _request):
        """
            request:
                -command_type
                -payload
                -response_channel
            response: (on response_channel)
                - RESULT
            * Send the payload on command_channel (self.namespace+"::command")
                ** redis-left-push (LPUSH)
            * Keep listening on response_channel (BLPOP)
        """
        assert isinstance(_request, dict)
        _request['response_channel'] = self._generate_response_channel()

        _redis = self.get_redis_connection()
        """
            The client always pushes in the left
            and the service always pushes in the right
        """
        if self.verbose:
            print("Request : ", _request)
        # Push request in command_channels
        # Note: The patched msgpack supports numpy arrays
        payload = msgpack.packb(_request, default=m.encode, use_bin_type=True)
        _redis.lpush(self.command_channel, payload)
        # Wait with a blocking pop for the response
        _response = _redis.blpop(_request['response_channel'])[1]
        if self.verbose:
            print("Response : ", _response)
        _response = msgpack.unpackb(
                        _response, 
                        object_hook=m.decode, 
                        encoding="utf8")
        if _response['type'] == messages.FLATLAND_RL.ERROR:
            raise Exception(str(_response["payload"]))
        else:
            return _response

    def ping_pong(self):
        """
            Official Handshake with the evaluation service
            Send a PING
            and wait for PONG
            If not PONG, raise error
        """
        _request = {}
        _request['type'] = messages.FLATLAND_RL.PING
        _request['payload'] = {}
        _response = self._blocking_request(_request)
        if _response['type'] != messages.FLATLAND_RL.PONG:
            raise Exception(
                "Unable to perform handshake with the redis service. \
                Expected PONG; received {}".format(json.dumps(_response)))
        else:
            return True

    def env_create(self, obs_builder_object):
        """
            Create a local env and remote env on which the 
            local agent can operate.
            The observation builder is only used in the local env
            and the remote env uses a DummyObservationBuilder
        """
        _request = {}
        _request['type'] = messages.FLATLAND_RL.ENV_CREATE
        _request['payload'] = {}
        _response = self._blocking_request(_request)
        observation = _response['payload']['observation']

        if not observation:
            # If the observation is False,
            # then the evaluations are complete
            # hence return false
            return observation

        test_env_file_path = _response['payload']['env_file_path']
        print("Received Env : ", test_env_file_path)
        test_env_file_path = os.path.join(
            self.test_envs_root,
            test_env_file_path
        )
        if not os.path.exists(test_env_file_path):
            raise Exception(
                "\nWe cannot seem to find the env file paths at the required location.\n"
                "Did you remember to set the AICROWD_TESTS_FOLDER environment variable "
                "to point to the location of the Tests folder ? \n"
                "We are currently looking at `{}` for the tests".format(self.test_envs_root)
                )
        print("Current env path : ", test_env_file_path)
        self.env = RailEnv(
            width=1,
            height=1,
            rail_generator=rail_from_file(test_env_file_path),
            obs_builder_object=obs_builder_object
        )
        self.env._max_episode_steps = \
            int(1.5 * (self.env.width + self.env.height))

        local_observation = self.env.reset()
        # Use the local observation 
        # as the remote server uses a dummy observation builder
        return local_observation

    def env_step(self, action, render=False):
        """
            Respond with [observation, reward, done, info]
        """
        _request = {}
        _request['type'] = messages.FLATLAND_RL.ENV_STEP
        _request['payload'] = {}
        _request['payload']['action'] = action
        _response = self._blocking_request(_request)
        _payload = _response['payload']
        
        # remote_observation = _payload['observation']
        remote_reward = _payload['reward']
        remote_done = _payload['done']
        remote_info = _payload['info']

        # Replicate the action in the local env
        local_observation, local_reward, local_done, local_info = \
            self.env.step(action)
        
        print(local_reward)
        if not are_dicts_equal(remote_reward, local_reward):
            raise Exception("local and remote `reward` are diverging")
            print(remote_reward, local_reward)
        if not are_dicts_equal(remote_done, local_done):
            raise Exception("local and remote `done` are diverging")
        
        # Return local_observation instead of remote_observation
        # as the remote_observation is build using a dummy observation
        # builder
        # We return the remote rewards and done as they are the 
        # once used by the evaluator
        return [local_observation, remote_reward, remote_done, remote_info]

    def submit(self):
        _request = {}
        _request['type'] = messages.FLATLAND_RL.ENV_SUBMIT
        _request['payload'] = {}
        _response = self._blocking_request(_request)
        if os.getenv("AICROWD_BLOCKING_SUBMIT"):
            """
            If the submission is supposed to happen as a blocking submit,
            then wait indefinitely for the evaluator to decide what to 
            do with the container.
            """
            while True:
                time.sleep(10)
        return _response['payload']


if __name__ == "__main__":
    remote_client = FlatlandRemoteClient()

    def my_controller(obs, _env):
        _action = {}
        for _idx, _ in enumerate(_env.agents):
            _action[_idx] = np.random.randint(0, 5)
        return _action
    
    my_observation_builder = TreeObsForRailEnv(max_depth=3,
                                predictor=ShortestPathPredictorForRailEnv())

    episode = 0
    obs = True
    while obs:        
        obs = remote_client.env_create(
                    obs_builder_object=my_observation_builder
                    )
        if not obs:
            """
            The remote env returns False as the first obs
            when it is done evaluating all the individual episodes
            """
            break
        print("Episode : {}".format(episode))
        episode += 1

        print(remote_client.env.dones['__all__'])

        while True:
            action = my_controller(obs, remote_client.env)
            observation, all_rewards, done, info = remote_client.env_step(action)
            if done['__all__']:
                print("Current Episode : ", episode)
                print("Episode Done")
                print("Reward : ", sum(list(all_rewards.values())))
                break

    print("Evaluation Complete...")       
    print(remote_client.submit())





from typing import Tuple

import numpy as np

from flatland.core.grid.grid4 import Grid4Transitions
from flatland.core.transition_map import GridTransitionMap


def make_simple_rail() -> Tuple[GridTransitionMap, np.array]:
    # We instantiate a very simple rail network on a 7x10 grid:
    #        |
    #        |
    #        |
    # _ _ _ /_\ _ _  _  _ _ _
    #               \ /
    #                |
    #                |
    #                |
    cells = [int('0000000000000000', 2),  # empty cell - Case 0
             int('1000000000100000', 2),  # Case 1 - straight
             int('1001001000100000', 2),  # Case 2 - simple switch
             int('1000010000100001', 2),  # Case 3 - diamond drossing
             int('1001011000100001', 2),  # Case 4 - single slip switch
             int('1100110000110011', 2),  # Case 5 - double slip switch
             int('0101001000000010', 2),  # Case 6 - symmetrical switch
             int('0010000000000000', 2)]  # Case 7 - dead end
    transitions = Grid4Transitions([])
    empty = cells[0]
    dead_end_from_south = cells[7]
    dead_end_from_west = transitions.rotate_transition(dead_end_from_south, 90)
    dead_end_from_north = transitions.rotate_transition(dead_end_from_south, 180)
    dead_end_from_east = transitions.rotate_transition(dead_end_from_south, 270)
    vertical_straight = cells[1]
    horizontal_straight = transitions.rotate_transition(vertical_straight, 90)
    double_switch_south_horizontal_straight = horizontal_straight + cells[6]
    double_switch_north_horizontal_straight = transitions.rotate_transition(
        double_switch_south_horizontal_straight, 180)
    rail_map = np.array(
        [[empty] * 3 + [dead_end_from_south] + [empty] * 6] +
        [[empty] * 3 + [vertical_straight] + [empty] * 6] * 2 +
        [[dead_end_from_east] + [horizontal_straight] * 2 +
         [double_switch_north_horizontal_straight] +
         [horizontal_straight] * 2 + [double_switch_south_horizontal_straight] +
         [horizontal_straight] * 2 + [dead_end_from_west]] +
        [[empty] * 6 + [vertical_straight] + [empty] * 3] * 2 +
        [[empty] * 6 + [dead_end_from_north] + [empty] * 3], dtype=np.uint16)
    rail = GridTransitionMap(width=rail_map.shape[1],
                             height=rail_map.shape[0], transitions=transitions)
    rail.grid = rail_map
    return rail, rail_map



import numpy as np

from flatland.core.grid.grid4 import Grid4Transitions
from flatland.core.transition_map import GridTransitionMap
from flatland.envs.generators import rail_from_grid_transition_map
from flatland.envs.observations import TreeObsForRailEnv
from flatland.envs.predictions import ShortestPathPredictorForRailEnv
from flatland.envs.rail_env import RailEnv


def test_walker():
    # _ _ _

    cells = [int('0000000000000000', 2),  # empty cell - Case 0
             int('1000000000100000', 2),  # Case 1 - straight
             int('1001001000100000', 2),  # Case 2 - simple switch
             int('1000010000100001', 2),  # Case 3 - diamond drossing
             int('1001011000100001', 2),  # Case 4 - single slip switch
             int('1100110000110011', 2),  # Case 5 - double slip switch
             int('0101001000000010', 2),  # Case 6 - symmetrical switch
             int('0010000000000000', 2)]  # Case 7 - dead end
    transitions = Grid4Transitions([])
    dead_end_from_south = cells[7]
    dead_end_from_west = transitions.rotate_transition(dead_end_from_south, 90)
    dead_end_from_east = transitions.rotate_transition(dead_end_from_south, 270)
    vertical_straight = cells[1]
    horizontal_straight = transitions.rotate_transition(vertical_straight, 90)

    rail_map = np.array(
        [[dead_end_from_east] + [horizontal_straight] + [dead_end_from_west]], dtype=np.uint16)
    rail = GridTransitionMap(width=rail_map.shape[1],
                             height=rail_map.shape[0], transitions=transitions)
    rail.grid = rail_map
    env = RailEnv(width=rail_map.shape[1],
                  height=rail_map.shape[0],
                  rail_generator=rail_from_grid_transition_map(rail),
                  number_of_agents=1,
                  obs_builder_object=TreeObsForRailEnv(max_depth=2,
                                                       predictor=ShortestPathPredictorForRailEnv(max_depth=10)),
                  )
    # reset to initialize agents_static
    env.reset()

    # set initial position and direction for testing...
    env.agents_static[0].position = (0, 1)
    env.agents_static[0].direction = 1
    env.agents_static[0].target = (0, 0)

    # reset to set agents from agents_static
    env.reset(False, False)
    obs_builder: TreeObsForRailEnv = env.obs_builder

    print(obs_builder.distance_map[(0, *[0, 1], 1)])
    assert obs_builder.distance_map[(0, *[0, 1], 1)] == 3
    print(obs_builder.distance_map[(0, *[0, 2], 3)])
    assert obs_builder.distance_map[(0, *[0, 2], 1)] == 2



import numpy as np

from flatland.envs.generators import complex_rail_generator
from flatland.envs.observations import TreeObsForRailEnv
from flatland.envs.predictions import ShortestPathPredictorForRailEnv
from flatland.envs.rail_env import RailEnv

np.random.seed(1)

# Use the complex_rail_generator to generate feasible network configurations with corresponding tasks
# Training on simple small tasks is the best way to get familiar with the environment
#

TreeObservation = TreeObsForRailEnv(max_depth=2, predictor=ShortestPathPredictorForRailEnv())
env = RailEnv(width=20,
              height=20,
              rail_generator=complex_rail_generator(nr_start_goal=10, nr_extra=1, min_dist=8, max_dist=99999, seed=0),
              obs_builder_object=TreeObservation,
              number_of_agents=2)


# Import your own Agent or use RLlib to train agents on Flatland
# As an example we use a random agent here


class RandomAgent:

    def __init__(self, state_size, action_size):
        self.state_size = state_size
        self.action_size = action_size

    def act(self, state):
        """
        :param state: input is the observation of the agent
        :return: returns an action
        """
        return np.random.choice(np.arange(self.action_size))

    def step(self, memories):
        """
        Step function to improve agent by adjusting policy given the observations

        :param memories: SARS Tuple to be
        :return:
        """
        return

    def save(self, filename):
        # Store the current policy
        return

    def load(self, filename):
        # Load a policy
        return


# Initialize the agent with the parameters corresponding to the environment and observation_builder
agent = RandomAgent(218, 4)
n_trials = 5

# Empty dictionary for all agent action
action_dict = dict()
print("Starting Training...")

for trials in range(1, n_trials + 1):

    # Reset environment and get initial observations for all agents
    obs = env.reset()
    # Here you can also further enhance the provided observation by means of normalization
    # See training navigation example in the baseline repository

    score = 0
    # Run episode
    for step in range(100):
        # Chose an action for each agent in the environment
        for a in range(env.get_num_agents()):
            action = agent.act(obs[a])
            action_dict.update({a: action})

        # Environment step which returns the observations for all agents, their corresponding
        # reward and whether their are done
        next_obs, all_rewards, done, _ = env.step(action_dict)
        # Update replay buffer and train agent
        for a in range(env.get_num_agents()):
            agent.step((obs[a], action_dict[a], all_rewards[a], next_obs[a], done[a]))
            score += all_rewards[a]

        obs = next_obs.copy()
        if done['__all__']:
            break
    print('Episode Nr. {}\t Score = {}'.format(trials, score))



import runpy
import sys
from io import StringIO

import importlib_resources
import pkg_resources
from importlib_resources import path

from benchmarks.benchmark_utils import swap_attr

for entry in [entry for entry in importlib_resources.contents('examples') if
              not pkg_resources.resource_isdir('examples', entry)
              and entry.endswith(".py")
              and '__init__' not in entry
              and 'demo.py' not in entry
              ]:
    with path('examples', entry) as file_in:
        print("")
        print("")
        print("")
        print("*****************************************************************")
        print("Running {}".format(entry))
        print("*****************************************************************")
        with swap_attr(sys, "stdin", StringIO("q")):
            runpy.run_path(file_in, run_name="__main__")



from enum import IntEnum

import numpy as np

from flatland.core.transitions import Transitions


class Grid8TransitionsEnum(IntEnum):
    NORTH = 0
    NORTH_EAST = 1
    EAST = 2
    SOUTH_EAST = 3
    SOUTH = 4
    SOUTH_WEST = 5
    WEST = 6
    NORTH_WEST = 7


class Grid8Transitions(Transitions):
    """
    Grid8Transitions class derived from Transitions.

    Special case of `Transitions' over a 2D-grid (FlatLand).
    Transitions are possible to neighboring cells on the grid if allowed.
    GridTransitions keeps track of valid transitions supplied as `transitions'
    list, each represented as a bitmap of 64 bits.

    0=North, 1=North-East, etc.

    """

    def __init__(self, transitions):
        self.transitions = transitions

    def get_type(self):
        return np.uint64

    def get_transitions(self, cell_transition, orientation):
        """
        Get the 8 possible transitions.

        Parameters
        ----------
        cell_transition : int
            64 bits used to encode the valid transitions for a cell.
        orientation : int
            Orientation of the agent inside the cell.

        Returns
        -------
        tuple
            List of the validity of transitions in the cell.

        """
        bits = (np.uint64(cell_transition) >> np.uint64((7 - orientation) * 8))
        cell_transition = (
            (bits >> np.uint64(7)) & np.uint64(1),
            (bits >> np.uint64(6)) & np.uint64(1),
            (bits >> np.uint64(5)) & np.uint64(1),
            (bits >> np.uint64(4)) & np.uint64(1),
            (bits >> np.uint64(3)) & np.uint64(1),
            (bits >> np.uint64(2)) & np.uint64(1),
            (bits >> np.uint64(1)) & np.uint64(1),
            bits & np.uint64(1))

        return cell_transition

    def set_transitions(self, cell_transition, orientation, new_transitions):
        """
        Set the possible transitions.

        Parameters
        ----------
        cell_transition : int
            64 bits used to encode the valid transitions for a cell.
        orientation : int
            Orientation of the agent inside the cell.
        new_transitions : tuple
            Tuple of new transitions validitiy for the cell.

        Returns
        -------
        int
            An updated bitmap that replaces the original transitions validity
            of `cell_transition' with `new_transitions', for the appropriate
            `orientation'.

        """
        mask = (1 << ((8 - orientation) * 8)) - (1 << ((7 - orientation) * 8))
        negmask = ~mask

        new_transitions = \
            (int(new_transitions[0]) & 1) << 7 | \
            (int(new_transitions[1]) & 1) << 6 | \
            (int(new_transitions[2]) & 1) << 5 | \
            (int(new_transitions[3]) & 1) << 4 | \
            (int(new_transitions[4]) & 1) << 3 | \
            (int(new_transitions[5]) & 1) << 2 | \
            (int(new_transitions[6]) & 1) << 1 | \
            (int(new_transitions[7]) & 1)

        cell_transition = (int(cell_transition) & negmask) | (new_transitions << ((7 - orientation) * 8))

        return cell_transition

    def get_transition(self, cell_transition, orientation, direction):
        """
        Get the transition bit (1 value) that determines whether an agent
        oriented in direction `orientation' and inside a cell with transitions
        `cell_transition' can move to the cell in direction `direction'
        relative to the current cell.

        Parameters
        ----------
        cell_transition : int
            64 bits used to encode the valid transitions for a cell.
        orientation : int
            Orientation of the agent inside the cell.
        direction : int
            Direction of movement whose validity is to be tested.

        Returns
        -------
        int
            Validity of the requested transition: 0/1 allowed/not allowed.

        """
        return ((cell_transition >> ((8 - 1 - orientation) * 8)) >> (8 - 1 - direction)) & 1

    def set_transition(self, cell_transition, orientation, direction, new_transition, remove_deadends=False):

        """
        Set the transition bit (1 value) that determines whether an agent
        oriented in direction `orientation' and inside a cell with transitions
        `cell_transition' can move to the cell in direction `direction'
        relative to the current cell.

        Parameters
        ----------
        cell_transition : int
            64 bits used to encode the valid transitions for a cell.
        orientation : int
            Orientation of the agent inside the cell.
        direction : int
            Direction of movement whose validity is to be tested.
        new_transition : int
            Validity of the requested transition: 0/1 allowed/not allowed.

        Returns
        -------
        int
            An updated bitmap that replaces the original transitions validity
            of `cell_transition' with `new_transitions', for the appropriate
            `orientation'.

        """
        if new_transition:
            cell_transition |= (1 << ((8 - 1 - orientation) * 8 + (8 - 1 - direction)))
        else:
            cell_transition &= ~(1 << ((8 - 1 - orientation) * 8 + (8 - 1 - direction)))

        return cell_transition

    def rotate_transition(self, cell_transition, rotation=0):
        """
        Clockwise-rotate a 64-bit transition bitmap by
        rotation={0, 45, 90, 135, 180, 225, 270, 315} degrees.

        Parameters
        ----------
        cell_transition : int
            64 bits used to encode the valid transitions for a cell.
        rotation : int
            Angle by which to clock-wise rotate the transition bits in
            `cell_transition' by. I.e., rotation={0, 45, 90, 135, 180,
            225, 270, 315} degrees.

        Returns
        -------
        int
            An updated bitmap that replaces the original transitions bits
            with the equivalent bitmap after rotation.

        """
        # TODO: WARNING: this part of the function has never been tested!

        # Rotate the individual bits in each block
        value = cell_transition
        rotation = rotation // 45
        for i in range(8):
            block_tuple = self.get_transitions(value, i)
            block_tuple = block_tuple[rotation:] + block_tuple[:rotation]
            value = self.set_transitions(value, i, block_tuple)

        # Rotate the 8bits blocks
        value = ((value & (2 ** (rotation * 8) - 1)) << ((8 - rotation) * 8)) | (value >> (rotation * 8))

        cell_transition = value

        return cell_transition

    def get_direction_enum(self) -> IntEnum:
        return Grid8TransitionsEnum



import random

import numpy as np

from flatland.core.grid.rail_env_grid import RailEnvTransitions
from flatland.core.transition_map import GridTransitionMap
from flatland.envs.rail_env import RailEnv
from flatland.utils.rendertools import RenderTool

random.seed(100)
np.random.seed(100)


def custom_rail_generator():
    def generator(width, height, num_agents=0, num_resets=0):
        rail_trans = RailEnvTransitions()
        grid_map = GridTransitionMap(width=width, height=height, transitions=rail_trans)
        rail_array = grid_map.grid
        rail_array.fill(0)
        new_tran = rail_trans.set_transition(1, 1, 1, 1)
        print(new_tran)
        agents_positions = []
        agents_direction = []
        agents_target = []
        rail_array[0, 0] = new_tran
        rail_array[0, 1] = new_tran
        return grid_map, agents_positions, agents_direction, agents_target

    return generator


env = RailEnv(width=6,
              height=4,
              rail_generator=custom_rail_generator(),
              number_of_agents=1)

env.reset()

env_renderer = RenderTool(env)
env_renderer.render_env(show=True)

input("Press Enter to continue...")



from itertools import starmap

import numpy as np
from attr import attrs, attrib, Factory


@attrs
class EnvAgentStatic(object):
    """ EnvAgentStatic - Stores initial position, direction and target.
        This is like static data for the environment - it's where an agent starts,
        rather than where it is at the moment.
        The target should also be stored here.
    """
    position = attrib()
    direction = attrib()
    target = attrib()
    moving = attrib(default=False)
    # speed_data: speed is added to position_fraction on each moving step, until position_fraction>=1.0,
    # after which 'transition_action_on_cellexit' is executed (equivalent to executing that action in the previous
    # cell if speed=1, as default)
    # N.B. we need to use factory since default arguments are not recreated on each call!
    speed_data = attrib(
        default=Factory(lambda: dict({'position_fraction': 0.0, 'speed': 1.0, 'transition_action_on_cellexit': 0})))

    @classmethod
    def from_lists(cls, positions, directions, targets, speeds=None):
        """ Create a list of EnvAgentStatics from lists of positions, directions and targets
        """
        speed_datas = []
        for i in range(len(positions)):
            speed_datas.append({'position_fraction': 0.0,
                                'speed': speeds[i] if speeds is not None else 1.0,
                                'transition_action_on_cellexit': 0})
        return list(starmap(EnvAgentStatic, zip(positions, directions, targets, [False] * len(positions), speed_datas)))

    def to_list(self):

        # I can't find an expression which works on both tuples, lists and ndarrays
        # which converts them all to a list of native python ints.
        lPos = self.position
        if type(lPos) is np.ndarray:
            lPos = lPos.tolist()

        lTarget = self.target
        if type(lTarget) is np.ndarray:
            lTarget = lTarget.tolist()

        return [lPos, int(self.direction), lTarget, int(self.moving), self.speed_data]


@attrs
class EnvAgent(EnvAgentStatic):
    """ EnvAgent - replace separate agent_* lists with a single list
        of agent objects.  The EnvAgent represent's the environment's view
        of the dynamic agent state.
        We are duplicating target in the EnvAgent, which seems simpler than
        forcing the env to refer to it in the EnvAgentStatic
    """
    handle = attrib(default=None)
    old_direction = attrib(default=None)
    old_position = attrib(default=None)

    def to_list(self):
        return [
            self.position, self.direction, self.target, self.handle,
            self.old_direction, self.old_position, self.moving, self.speed_data]

    @classmethod
    def from_static(cls, oStatic):
        """ Create an EnvAgent from the EnvAgentStatic,
        copying all the fields, and adding handle with the default 0.
        """
        return EnvAgent(*oStatic.__dict__, handle=0)

    @classmethod
    def list_from_static(cls, lEnvAgentStatic, handles=None):
        """ Create an EnvAgent from the EnvAgentStatic,
        copying all the fields, and adding handle with the default 0.
        """
        if handles is None:
            handles = range(len(lEnvAgentStatic))

        return [EnvAgent(**oEAS.__dict__, handle=handle)
                for handle, oEAS in zip(handles, lEnvAgentStatic)]



import shlex
import sys
from subprocess import Popen, PIPE

import importlib_resources
import pkg_resources
from importlib_resources import path
from ipython_genutils.py3compat import string_types, bytes_to_str


# taken from https://github.com/jupyter/nbconvert/blob/master/nbconvert/tests/base.py
def run_python(parameters, ignore_return_code=False, stdin=None):
    """
    Run python as a shell command, listening for both Errors and
    non-zero return codes. Returns the tuple (stdout, stderr) of
    output produced during the nbconvert run.
    Parameters
    ----------
    parameters : str, list(str)
        List of parameters to pass to IPython.
    ignore_return_code : optional bool (default False)
        Throw an OSError if the return code
    """
    cmd = [sys.executable]
    if sys.platform == 'win32':
        if isinstance(parameters, string_types):
            cmd = ' '.join(cmd) + ' ' + parameters
        else:
            cmd = ' '.join(cmd + parameters)
    else:
        if isinstance(parameters, string_types):
            parameters = shlex.split(parameters)
        cmd += parameters
    p = Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE)
    stdout, stderr = p.communicate(input=stdin)
    if not (p.returncode == 0 or ignore_return_code):
        raise OSError(bytes_to_str(stderr))
    return stdout.decode('utf8', 'replace'), stderr.decode('utf8', 'replace')


for entry in [entry for entry in importlib_resources.contents('notebooks') if
              not pkg_resources.resource_isdir('notebooks', entry)
              and entry.endswith(".ipynb")
              ]:
    print("*****************************************************************")
    print("Converting and running {}".format(entry))
    print("*****************************************************************")

    with path('notebooks', entry) as file_in:
        out, err = run_python(" -m jupyter nbconvert --execute --to notebook --inplace " + str(file_in))
        sys.stderr.write(err)
        sys.stderr.flush()
        sys.stdout.write(out)
        sys.stdout.flush()



import os
import boto3
import uuid
import subprocess

###############################################################
# Expected Env Variables
###############################################################
# Default Values to be provided :
# AICROWD_IS_GRADING : true
# CROWDAI_IS_GRADING : true
# S3_BUCKET : aicrowd-production
# S3_UPLOAD_PATH_TEMPLATE : misc/flatland-rl-Media/{}.mp4
# AWS_ACCESS_KEY_ID
# AWS_SECRET_ACCESS_KEY
# http_proxy
# https_proxy
###############################################################
AWS_ACCESS_KEY_ID = os.getenv("AWS_ACCESS_KEY_ID", False)
AWS_SECRET_ACCESS_KEY = os.getenv("AWS_SECRET_ACCESS_KEY", False)
S3_BUCKET = os.getenv("S3_BUCKET", "aicrowd-production")
S3_UPLOAD_PATH_TEMPLATE = os.getenv("S3_UPLOAD_PATH_TEMPLATE", "misc/flatland-rl-Media/{}.mp4")


def get_boto_client():
    if not AWS_ACCESS_KEY_ID or not AWS_SECRET_ACCESS_KEY:
        raise Exception("AWS Credentials not provided..")
    return boto3.client(
            's3',
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY
    )


def is_aws_configured():
    if not AWS_ACCESS_KEY_ID or not AWS_SECRET_ACCESS_KEY:
        return False
    else:
        return True


def is_grading():
    return os.getenv("CROWDAI_IS_GRADING", False) or \
        os.getenv("AICROWD_IS_GRADING", False)


def upload_to_s3(localpath):
    s3 = get_boto_client()
    if not S3_UPLOAD_PATH_TEMPLATE:
        raise Exception("S3_UPLOAD_PATH_TEMPLATE not provided...")
    if not S3_BUCKET:
        raise Exception("S3_BUCKET not provided...")
    
    image_target_key = S3_UPLOAD_PATH_TEMPLATE.format(str(uuid.uuid4()))
    s3.put_object(
        ACL="public-read",
        Bucket=S3_BUCKET,
        Key=image_target_key,
        Body=open(localpath, 'rb')
    )
    return image_target_key


def make_subprocess_call(command, shell=False):
    result = subprocess.run(
                command.split(),
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
    stdout = result.stdout.decode('utf-8')
    stderr = result.stderr.decode('utf-8')
    return result.returncode, stdout, stderr


def generate_movie_from_frames(frames_folder):
    """
        Expects the frames in the  frames_folder folder 
        and then use ffmpeg to generate the video
        which writes the output to the frames_folder
    """
    # Generate Thumbnail Video
    print("Generating Thumbnail...")
    frames_path = os.path.join(frames_folder, "flatland_frame_%04d.png")
    thumb_output_path = os.path.join(frames_folder, "out_thumb.mp4")
    return_code, output, output_err = make_subprocess_call(
        "ffmpeg -r 7 -start_number 0 -i " + 
        frames_path + 
        " -c:v libx264 -vf fps=7 -pix_fmt yuv420p -s 320x320 " + 
        thumb_output_path
    )
    if return_code != 0:
        raise Exception(output_err)

    # Generate Normal Sized Video
    print("Generating Normal Video...")
    frames_path = os.path.join(frames_folder, "flatland_frame_%04d.png")
    output_path = os.path.join(frames_folder, "out.mp4")
    return_code, output, output_err = make_subprocess_call(
        "ffmpeg -r 7 -start_number 0 -i " + 
        frames_path + 
        " -c:v libx264 -vf fps=7 -pix_fmt yuv420p -s 600x600 " + 
        output_path
    )
    if return_code != 0:
        raise Exception(output_err)

    return output_path, thumb_output_path




import contextlib


# Copied from https://docs.python.org/3/library/test.html

@contextlib.contextmanager
def swap_attr(obj, attr, new_val):
    """Temporary swap out an attribute with a new object.

    Usage:
        with swap_attr(obj, "attr", 5):
            ...

        This will set obj.attr to 5 for the duration of the with: block,
        restoring the old value at the end of the block. If `attr` doesn't
        exist on `obj`, it will be created and then deleted at the end of the
        block.

        The old value (or None if it doesn't exist) will be assigned to the
        target of the "as" clause, if there is one.
    """
    if hasattr(obj, attr):
        real_val = getattr(obj, attr)
        setattr(obj, attr, new_val)
        try:
            yield real_val
        finally:
            setattr(obj, attr, real_val)
    else:
        setattr(obj, attr, new_val)
        try:
            yield
        finally:
            if hasattr(obj, attr):
                delattr(obj, attr)



"""
Definition of the RailEnv environment.
"""
# TODO:  _ this is a global method --> utils or remove later

from enum import IntEnum

import msgpack
import msgpack_numpy as m
import numpy as np

from flatland.core.env import Environment
from flatland.core.grid.grid4_utils import get_new_position
from flatland.envs.agent_utils import EnvAgentStatic, EnvAgent
from flatland.envs.generators import random_rail_generator
from flatland.envs.observations import TreeObsForRailEnv

m.patch()


class RailEnvActions(IntEnum):
    DO_NOTHING = 0  # implies change of direction in a dead-end!
    MOVE_LEFT = 1
    MOVE_FORWARD = 2
    MOVE_RIGHT = 3
    STOP_MOVING = 4

    @staticmethod
    def to_char(a: int):
        return {
            0: 'B',
            1: 'L',
            2: 'F',
            3: 'R',
            4: 'S',
        }[a]


class RailEnv(Environment):
    """
    RailEnv environment class.

    RailEnv is an environment inspired by a (simplified version of) a rail
    network, in which agents (trains) have to navigate to their target
    locations in the shortest time possible, while at the same time cooperating
    to avoid bottlenecks.

    The valid actions in the environment are:

     -   0: do nothing (continue moving or stay still)
     -   1: turn left at switch and move to the next cell; if the agent was not moving, movement is started
     -   2: move to the next cell in front of the agent; if the agent was not moving, movement is started
     -   3: turn right at switch and move to the next cell; if the agent was not moving, movement is started
     -   4: stop moving

    Moving forward in a dead-end cell makes the agent turn 180 degrees and step
    to the cell it came from.


    The actions of the agents are executed in order of their handle to prevent
    deadlocks and to allow them to learn relative priorities.

    Reward Function:

    It costs each agent a step_penalty for every time-step taken in the environment. Independent of the movement
    of the agent. Currently all other penalties such as penalty for stopping, starting and invalid actions are set to 0.

    alpha = 1
    beta = 1
    Reward function parameters:

    - invalid_action_penalty = 0
    - step_penalty = -alpha
    - global_reward = beta
    - stop_penalty = 0  # penalty for stopping a moving agent
    - start_penalty = 0  # penalty for starting a stopped agent

    """

    def __init__(self,
                 width,
                 height,
                 rail_generator=random_rail_generator(),
                 number_of_agents=1,
                 obs_builder_object=TreeObsForRailEnv(max_depth=2),
                 max_episode_steps=None
                 ):
        """
        Environment init.

        Parameters
        -------
        rail_generator : function
            The rail_generator function is a function that takes the width,
            height and agents handles of a  rail environment, along with the number of times
            the env has been reset, and returns a GridTransitionMap object and a list of
            starting positions, targets, and initial orientations for agent handle.
            Implemented functions are:
                random_rail_generator : generate a random rail of given size
                rail_from_grid_transition_map(rail_map) : generate a rail from
                                        a GridTransitionMap object
                rail_from_manual_sp ecifications_generator(rail_spec) : generate a rail from
                                        a rail specifications array
                TODO: generate_rail_from_saved_list or from list of ndarray bitmaps ---
        width : int
            The width of the rail map. Potentially in the future,
            a range of widths to sample from.
        height : int
            The height of the rail map. Potentially in the future,
            a range of heights to sample from.
        number_of_agents : int
            Number of agents to spawn on the map. Potentially in the future,
            a range of number of agents to sample from.
        obs_builder_object: ObservationBuilder object
            ObservationBuilder-derived object that takes builds observation
            vectors for each agent.
        max_episode_steps : int or None

        file_name: you can load a pickle file.
        """

        self.rail_generator = rail_generator
        self.rail = None
        self.width = width
        self.height = height

        self.rewards = [0] * number_of_agents
        self.done = False
        self.obs_builder = obs_builder_object
        self.obs_builder._set_env(self)

        self._max_episode_steps = max_episode_steps
        self._elapsed_steps = 0

        self.dones = dict.fromkeys(list(range(number_of_agents)) + ["__all__"], False)

        self.obs_dict = {}
        self.rewards_dict = {}
        self.dev_obs_dict = {}
        self.dev_pred_dict = {}

        self.agents = [None] * number_of_agents  # live agents
        self.agents_static = [None] * number_of_agents  # static agent information
        self.num_resets = 0

        self.action_space = [1]
        self.observation_space = self.obs_builder.observation_space  # updated on resets?

        self.reset()
        self.num_resets = 0  # yes, set it to zero again!

        self.valid_positions = None

    # no more agent_handles
    def get_agent_handles(self):
        return range(self.get_num_agents())

    def get_num_agents(self, static=True):
        if static:
            return len(self.agents_static)
        else:
            return len(self.agents)

    def add_agent_static(self, agent_static):
        """ Add static info for a single agent.
            Returns the index of the new agent.
        """
        self.agents_static.append(agent_static)
        return len(self.agents_static) - 1

    def restart_agents(self):
        """ Reset the agents to their starting positions defined in agents_static
        """
        self.agents = EnvAgent.list_from_static(self.agents_static)

    def reset(self, regen_rail=True, replace_agents=True):
        """ if regen_rail then regenerate the rails.
            if replace_agents then regenerate the agents static.
            Relies on the rail_generator returning agent_static lists (pos, dir, target)
        """
        tRailAgents = self.rail_generator(self.width, self.height, self.get_num_agents(), self.num_resets)

        # Check if generator provided a distance map TODO: Make this check safer!
        if len(tRailAgents) > 5:
            self.obs_builder.distance_map = tRailAgents[-1]

        if regen_rail or self.rail is None:
            self.rail = tRailAgents[0]
            self.height, self.width = self.rail.grid.shape

        if replace_agents:
            self.agents_static = EnvAgentStatic.from_lists(*tRailAgents[1:5])

        self.restart_agents()

        for i_agent in range(self.get_num_agents()):
            agent = self.agents[i_agent]
            agent.speed_data['position_fraction'] = 0.0

        self.num_resets += 1
        self._elapsed_steps = 0

        # TODO perhaps dones should be part of each agent.
        self.dones = dict.fromkeys(list(range(self.get_num_agents())) + ["__all__"], False)

        # Reset the state of the observation builder with the new environment
        self.obs_builder.reset()
        self.observation_space = self.obs_builder.observation_space  # <-- change on reset?

        # Return the new observation vectors for each agent
        return self._get_observations()

    def step(self, action_dict_):
        self._elapsed_steps += 1

        action_dict = action_dict_.copy()

        alpha = 1.0
        beta = 1.0

        invalid_action_penalty = 0  # previously -2; GIACOMO: we decided that invalid actions will carry no penalty
        step_penalty = -1 * alpha
        global_reward = 1 * beta
        stop_penalty = 0  # penalty for stopping a moving agent
        start_penalty = 0  # penalty for starting a stopped agent

        # Reset the step rewards
        self.rewards_dict = dict()
        for i_agent in range(self.get_num_agents()):
            self.rewards_dict[i_agent] = 0

        if self.dones["__all__"]:
            self.rewards_dict = {i: r + global_reward for i, r in self.rewards_dict.items()}
            return self._get_observations(), self.rewards_dict, self.dones, {}

        # for i in range(len(self.agents_handles)):
        for i_agent in range(self.get_num_agents()):
            agent = self.agents[i_agent]
            agent.old_direction = agent.direction
            agent.old_position = agent.position
            if self.dones[i_agent]:  # this agent has already completed...
                continue

            if i_agent not in action_dict:  # no action has been supplied for this agent
                action_dict[i_agent] = RailEnvActions.DO_NOTHING

            if action_dict[i_agent] < 0 or action_dict[i_agent] > len(RailEnvActions):
                print('ERROR: illegal action=', action_dict[i_agent],
                      'for agent with index=', i_agent,
                      '"DO NOTHING" will be executed instead')
                action_dict[i_agent] = RailEnvActions.DO_NOTHING

            action = action_dict[i_agent]

            if action == RailEnvActions.DO_NOTHING and agent.moving:
                # Keep moving
                action = RailEnvActions.MOVE_FORWARD

            if action == RailEnvActions.STOP_MOVING and agent.moving and agent.speed_data['position_fraction'] == 0.:
                # Only allow halting an agent on entering new cells.
                agent.moving = False
                self.rewards_dict[i_agent] += stop_penalty

            if not agent.moving and not (action == RailEnvActions.DO_NOTHING or action == RailEnvActions.STOP_MOVING):
                # Allow agent to start with any forward or direction action
                agent.moving = True
                self.rewards_dict[i_agent] += start_penalty

            # Now perform a movement.
            # If the agent is in an initial position within a new cell (agent.speed_data['position_fraction']<eps)
            #   store the desired action in `transition_action_on_cellexit' (only if the desired transition is
            #   allowed! otherwise DO_NOTHING!)
            # Then in any case (if agent.moving) and the `transition_action_on_cellexit' is valid, increment the
            #   position_fraction by the speed of the agent   (regardless of action taken, as long as no
            #   STOP_MOVING, but that makes agent.moving=False)
            # If the new position fraction is >= 1, reset to 0, and perform the stored
            #   transition_action_on_cellexit

            # If the agent can make an action
            action_selected = False
            if agent.speed_data['position_fraction'] == 0.:
                if action != RailEnvActions.DO_NOTHING and action != RailEnvActions.STOP_MOVING:
                    cell_free, new_cell_valid, new_direction, new_position, transition_valid = \
                        self._check_action_on_agent(action, agent)

                    if all([new_cell_valid, transition_valid]):
                        agent.speed_data['transition_action_on_cellexit'] = action
                        action_selected = True

                    else:
                        # But, if the chosen invalid action was LEFT/RIGHT, and the agent is moving,
                        # try to keep moving forward!
                        if (action == RailEnvActions.MOVE_LEFT or action == RailEnvActions.MOVE_RIGHT) and agent.moving:
                            cell_free, new_cell_valid, new_direction, new_position, transition_valid = \
                                self._check_action_on_agent(RailEnvActions.MOVE_FORWARD, agent)

                            if all([new_cell_valid, transition_valid]):
                                agent.speed_data['transition_action_on_cellexit'] = RailEnvActions.MOVE_FORWARD
                                action_selected = True

                            else:
                                # TODO: an invalid action was chosen after entering the cell. The agent cannot move.
                                self.rewards_dict[i_agent] += invalid_action_penalty
                                self.rewards_dict[i_agent] += step_penalty * agent.speed_data['speed']
                                self.rewards_dict[i_agent] += stop_penalty
                                agent.moving = False
                                continue
                        else:
                            # TODO: an invalid action was chosen after entering the cell. The agent cannot move.
                            self.rewards_dict[i_agent] += invalid_action_penalty
                            self.rewards_dict[i_agent] += step_penalty * agent.speed_data['speed']
                            self.rewards_dict[i_agent] += stop_penalty
                            agent.moving = False
                            continue

            if agent.moving and (action_selected or agent.speed_data['position_fraction'] > 0.0):
                agent.speed_data['position_fraction'] += agent.speed_data['speed']

            if agent.speed_data['position_fraction'] >= 1.0:

                # Perform stored action to transition to the next cell

                cell_free, new_cell_valid, new_direction, new_position, transition_valid = \
                    self._check_action_on_agent(agent.speed_data['transition_action_on_cellexit'], agent)

                # Check that everything is still fee and that the agent can move
                if all([new_cell_valid, transition_valid, cell_free]):
                    agent.position = new_position
                    agent.direction = new_direction
                    agent.speed_data['position_fraction'] = 0.0
                else:
                    # If the agent cannot move due to any reason, we set its state to not moving.
                    agent.moving = False

            if np.equal(agent.position, agent.target).all():
                self.dones[i_agent] = True
                agent.moving = False
            else:
                self.rewards_dict[i_agent] += step_penalty * agent.speed_data['speed']

        # Check for end of episode + add global reward to all rewards!
        if np.all([np.array_equal(agent2.position, agent2.target) for agent2 in self.agents]):
            self.dones["__all__"] = True
            self.rewards_dict = {i: 0 * r + global_reward for i, r in self.rewards_dict.items()}

        if (self._max_episode_steps is not None) and (self._elapsed_steps >= self._max_episode_steps):
            self.dones["__all__"] = True
            for k in self.dones.keys():
                self.dones[k] = True

        return self._get_observations(), self.rewards_dict, self.dones, {}

    def _check_action_on_agent(self, action, agent):
        # compute number of possible transitions in the current
        # cell used to check for invalid actions
        new_direction, transition_valid = self.check_action(agent, action)
        new_position = get_new_position(agent.position, new_direction)

        # Is it a legal move?
        # 1) transition allows the new_direction in the cell,
        # 2) the new cell is not empty (case 0),
        # 3) the cell is free, i.e., no agent is currently in that cell
        new_cell_valid = (
            np.array_equal(  # Check the new position is still in the grid
                new_position,
                np.clip(new_position, [0, 0], [self.height - 1, self.width - 1]))
            and  # check the new position has some transitions (ie is not an empty cell)
            self.rail.get_full_transitions(*new_position) > 0)

        # If transition validity hasn't been checked yet.
        if transition_valid is None:
            transition_valid = self.rail.get_transition(
                (*agent.position, agent.direction),
                new_direction)

        # Check the new position is not the same as any of the existing agent positions
        # (including itself, for simplicity, since it is moving)
        cell_free = not np.any(
            np.equal(new_position, [agent2.position for agent2 in self.agents]).all(1))
        return cell_free, new_cell_valid, new_direction, new_position, transition_valid

    def check_action(self, agent, action):
        transition_valid = None
        possible_transitions = self.rail.get_transitions(*agent.position, agent.direction)
        num_transitions = np.count_nonzero(possible_transitions)

        new_direction = agent.direction
        if action == RailEnvActions.MOVE_LEFT:
            new_direction = agent.direction - 1
            if num_transitions <= 1:
                transition_valid = False

        elif action == RailEnvActions.MOVE_RIGHT:
            new_direction = agent.direction + 1
            if num_transitions <= 1:
                transition_valid = False

        new_direction %= 4

        if action == RailEnvActions.MOVE_FORWARD:
            if num_transitions == 1:
                # - dead-end, straight line or curved line;
                # new_direction will be the only valid transition
                # - take only available transition
                new_direction = np.argmax(possible_transitions)
                transition_valid = True
        return new_direction, transition_valid

    def _get_observations(self):
        self.obs_dict = self.obs_builder.get_many(list(range(self.get_num_agents())))
        return self.obs_dict

    def get_full_state_msg(self):
        grid_data = self.rail.grid.tolist()
        agent_static_data = [agent.to_list() for agent in self.agents_static]
        agent_data = [agent.to_list() for agent in self.agents]

        msgpack.packb(grid_data)
        msgpack.packb(agent_data)
        msgpack.packb(agent_static_data)

        msg_data = {
            "grid": grid_data,
            "agents_static": agent_static_data,
            "agents": agent_data}
        return msgpack.packb(msg_data, use_bin_type=True)

    def get_agent_state_msg(self):
        agent_data = [agent.to_list() for agent in self.agents]
        msg_data = {
            "agents": agent_data}
        return msgpack.packb(msg_data, use_bin_type=True)

    def set_full_state_msg(self, msg_data):
        data = msgpack.unpackb(msg_data, use_list=False)
        self.rail.grid = np.array(data[b"grid"])
        # agents are always reset as not moving
        self.agents_static = [EnvAgentStatic(d[0], d[1], d[2], moving=False) for d in data[b"agents_static"]]
        self.agents = [EnvAgent(d[0], d[1], d[2], d[3], d[4]) for d in data[b"agents"]]
        # setup with loaded data
        self.height, self.width = self.rail.grid.shape
        self.rail.height = self.height
        self.rail.width = self.width
        self.dones = dict.fromkeys(list(range(self.get_num_agents())) + ["__all__"], False)

    def set_full_state_dist_msg(self, msg_data):
        data = msgpack.unpackb(msg_data, use_list=False)
        self.rail.grid = np.array(data[b"grid"])
        # agents are always reset as not moving
        self.agents_static = [EnvAgentStatic(d[0], d[1], d[2], moving=False) for d in data[b"agents_static"]]
        self.agents = [EnvAgent(d[0], d[1], d[2], d[3], d[4]) for d in data[b"agents"]]
        if hasattr(self.obs_builder, 'distance_map') and b"distance_maps" in data.keys():
            self.obs_builder.distance_map = data[b"distance_maps"]
        # setup with loaded data
        self.height, self.width = self.rail.grid.shape
        self.rail.height = self.height
        self.rail.width = self.width
        self.dones = dict.fromkeys(list(range(self.get_num_agents())) + ["__all__"], False)

    def get_full_state_dist_msg(self):
        grid_data = self.rail.grid.tolist()
        agent_static_data = [agent.to_list() for agent in self.agents_static]
        agent_data = [agent.to_list() for agent in self.agents]

        msgpack.packb(grid_data)
        msgpack.packb(agent_data)
        msgpack.packb(agent_static_data)
        if hasattr(self.obs_builder, 'distance_map'):
            distance_map_data = self.obs_builder.distance_map
            msgpack.packb(distance_map_data)
            msg_data = {
                "grid": grid_data,
                "agents_static": agent_static_data,
                "agents": agent_data,
                "distance_maps": distance_map_data}
        else:
            msg_data = {
                "grid": grid_data,
                "agents_static": agent_static_data,
                "agents": agent_data}

        return msgpack.packb(msg_data, use_bin_type=True)

    def save(self, filename):
        if hasattr(self.obs_builder, 'distance_map'):
            if len(self.obs_builder.distance_map) > 0:
                with open(filename, "wb") as file_out:
                    file_out.write(self.get_full_state_dist_msg())
            else:
                with open(filename, "wb") as file_out:
                    file_out.write(self.get_full_state_msg())
        else:
            with open(filename, "wb") as file_out:
                file_out.write(self.get_full_state_msg())

    def load(self, filename):
        if hasattr(self.obs_builder, 'distance_map'):
            with open(filename, "rb") as file_in:
                load_data = file_in.read()
                self.set_full_state_dist_msg(load_data)
        else:
            with open(filename, "rb") as file_in:
                load_data = file_in.read()
                self.set_full_state_msg(load_data)

    def load_pkl(self, pkl_data):
        self.set_full_state_msg(pkl_data)

    def load_resource(self, package, resource):
        from importlib_resources import read_binary
        load_data = read_binary(package, resource)
        self.set_full_state_msg(load_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np

from flatland.core.grid.grid4 import Grid4Transitions
from flatland.core.grid.rail_env_grid import RailEnvTransitions
from flatland.core.transition_map import GridTransitionMap
from flatland.envs.agent_utils import EnvAgent
from flatland.envs.agent_utils import EnvAgentStatic
from flatland.envs.generators import complex_rail_generator
from flatland.envs.generators import rail_from_grid_transition_map
from flatland.envs.observations import GlobalObsForRailEnv
from flatland.envs.rail_env import RailEnv

"""Tests for `flatland` package."""


def test_load_env():
    env = RailEnv(10, 10)
    env.load_resource('env_data.tests', 'test-10x10.mpk')

    agent_static = EnvAgentStatic((0, 0), 2, (5, 5), False)
    env.add_agent_static(agent_static)
    assert env.get_num_agents() == 1


def test_save_load():
    env = RailEnv(width=10, height=10,
                  rail_generator=complex_rail_generator(nr_start_goal=2, nr_extra=5, min_dist=6, seed=0),
                  number_of_agents=2)
    env.reset()
    agent_1_pos = env.agents_static[0].position
    agent_1_dir = env.agents_static[0].direction
    agent_1_tar = env.agents_static[0].target
    agent_2_pos = env.agents_static[1].position
    agent_2_dir = env.agents_static[1].direction
    agent_2_tar = env.agents_static[1].target
    env.save("test_save.dat")
    env.load("test_save.dat")
    assert (env.width == 10)
    assert (env.height == 10)
    assert (len(env.agents) == 2)
    assert (agent_1_pos == env.agents_static[0].position)
    assert (agent_1_dir == env.agents_static[0].direction)
    assert (agent_1_tar == env.agents_static[0].target)
    assert (agent_2_pos == env.agents_static[1].position)
    assert (agent_2_dir == env.agents_static[1].direction)
    assert (agent_2_tar == env.agents_static[1].target)


def test_rail_environment_single_agent():
    cells = [int('0000000000000000', 2),  # empty cell - Case 0
             int('1000000000100000', 2),  # Case 1 - straight
             int('1001001000100000', 2),  # Case 2 - simple switch
             int('1000010000100001', 2),  # Case 3 - diamond drossing
             int('1001011000100001', 2),  # Case 4 - single slip switch
             int('1100110000110011', 2),  # Case 5 - double slip switch
             int('0101001000000010', 2),  # Case 6 - symmetrical switch
             int('0010000000000000', 2)]  # Case 7 - dead end

    # We instantiate the following map on a 3x3 grid
    #  _  _
    # / \/ \
    # | |  |
    # \_/\_/

    transitions = RailEnvTransitions()
    vertical_line = cells[1]
    south_symmetrical_switch = cells[6]
    north_symmetrical_switch = transitions.rotate_transition(south_symmetrical_switch, 180)
    # Simple turn not in the base transitions ?
    south_east_turn = int('0100000000000010', 2)
    south_west_turn = transitions.rotate_transition(south_east_turn, 90)
    north_east_turn = transitions.rotate_transition(south_east_turn, 270)
    north_west_turn = transitions.rotate_transition(south_east_turn, 180)

    rail_map = np.array([[south_east_turn, south_symmetrical_switch,
                          south_west_turn],
                         [vertical_line, vertical_line, vertical_line],
                         [north_east_turn, north_symmetrical_switch,
                          north_west_turn]],
                        dtype=np.uint16)

    rail = GridTransitionMap(width=3, height=3, transitions=transitions)
    rail.grid = rail_map
    rail_env = RailEnv(width=3,
                       height=3,
                       rail_generator=rail_from_grid_transition_map(rail),
                       number_of_agents=1,
                       obs_builder_object=GlobalObsForRailEnv())

    for _ in range(200):
        _ = rail_env.reset()

        # We do not care about target for the moment
        agent = rail_env.agents[0]
        agent.target = [-1, -1]

        # Check that trains are always initialized at a consistent position
        # or direction.
        # They should always be able to go somewhere.
        assert (transitions.get_transitions(
            rail_map[agent.position],
            agent.direction) != (0, 0, 0, 0))

        initial_pos = agent.position

        valid_active_actions_done = 0
        pos = initial_pos
        while valid_active_actions_done < 6:
            # We randomly select an action
            action = np.random.randint(4)

            _, _, _, _ = rail_env.step({0: action})

            prev_pos = pos
            pos = agent.position  # rail_env.agents_position[0]
            if prev_pos != pos:
                valid_active_actions_done += 1

        # After 6 movements on this railway network, the train should be back
        # to its original height on the map.
        assert (initial_pos[0] == agent.position[0])

        # We check that the train always attains its target after some time
        for _ in range(10):
            _ = rail_env.reset()

            done = False
            while not done:
                # We randomly select an action
                action = np.random.randint(4)

                _, _, dones, _ = rail_env.step({0: action})
                done = dones['__all__']


test_rail_environment_single_agent()


def test_dead_end():
    transitions = Grid4Transitions([])

    straight_vertical = int('1000000000100000', 2)  # Case 1 - straight
    straight_horizontal = transitions.rotate_transition(straight_vertical,
                                                        90)

    dead_end_from_south = int('0010000000000000', 2)  # Case 7 - dead end

    # We instantiate the following railway
    # O->-- where > is the train and O the target. After 6 steps,
    # the train should be done.

    rail_map = np.array(
        [[transitions.rotate_transition(dead_end_from_south, 270)] +
         [straight_horizontal] * 3 +
         [transitions.rotate_transition(dead_end_from_south, 90)]],
        dtype=np.uint16)

    rail = GridTransitionMap(width=rail_map.shape[1],
                             height=rail_map.shape[0],
                             transitions=transitions)

    rail.grid = rail_map
    rail_env = RailEnv(width=rail_map.shape[1],
                       height=rail_map.shape[0],
                       rail_generator=rail_from_grid_transition_map(rail),
                       number_of_agents=1,
                       obs_builder_object=GlobalObsForRailEnv())

    def check_consistency(rail_env):
        # We run step to check that trains do not move anymore
        # after being done.
        # TODO: GIACOMO: this is deprecated and should be updated; thenew behavior is that agents keep moving
        # until they are manually stopped.
        for i in range(7):
            prev_pos = rail_env.agents[0].position

            # The train cannot turn, so we check that when it tries,
            # it stays where it is.
            _ = rail_env.step({0: 1})
            _ = rail_env.step({0: 3})
            assert (rail_env.agents[0].position == prev_pos)
            _, _, dones, _ = rail_env.step({0: 2})

            if i < 5:
                assert (not dones[0] and not dones['__all__'])
            else:
                assert (dones[0] and dones['__all__'])

    # We try the configuration in the 4 directions:
    rail_env.reset()
    rail_env.agents = [EnvAgent(position=(0, 2), direction=1, target=(0, 0), moving=False)]

    rail_env.reset()
    rail_env.agents = [EnvAgent(position=(0, 2), direction=3, target=(0, 4), moving=False)]

    # In the vertical configuration:
    rail_map = np.array(
        [[dead_end_from_south]] + [[straight_vertical]] * 3 +
        [[transitions.rotate_transition(dead_end_from_south, 180)]],
        dtype=np.uint16)

    rail = GridTransitionMap(width=rail_map.shape[1],
                             height=rail_map.shape[0],
                             transitions=transitions)

    rail.grid = rail_map
    rail_env = RailEnv(width=rail_map.shape[1],
                       height=rail_map.shape[0],
                       rail_generator=rail_from_grid_transition_map(rail),
                       number_of_agents=1,
                       obs_builder_object=GlobalObsForRailEnv())

    rail_env.reset()
    rail_env.agents = [EnvAgent(position=(2, 0), direction=2, target=(0, 0), moving=False)]

    rail_env.reset()
    rail_env.agents = [EnvAgent(position=(2, 0), direction=0, target=(4, 0), moving=False)]



import time
from collections import deque
from enum import IntEnum

import numpy as np
from numpy import array
from recordtype import recordtype

from flatland.utils.graphics_pil import PILGL, PILSVG


# TODO: suggested renaming to RailEnvRenderTool, as it will only work with RailEnv!

class AgentRenderVariant(IntEnum):
    BOX_ONLY = 0
    ONE_STEP_BEHIND = 1
    AGENT_SHOWS_OPTIONS = 2
    ONE_STEP_BEHIND_AND_BOX = 3
    AGENT_SHOWS_OPTIONS_AND_BOX = 4


class RenderTool(object):
    """ Class to render the RailEnv and agents.
        Uses two layers, layer 0 for rails (mostly static), layer 1 for agents etc (dynamic)
        The lower / rail layer 0 is only redrawn after set_new_rail() has been called.
        Created with a "GraphicsLayer" or gl - now either PIL or PILSVG
    """
    visit = recordtype("visit", ["rc", "iDir", "iDepth", "prev"])

    color_list = list("brgcmyk")
    # \delta RC for NESW
    transitions_row_col = np.array([[-1, 0], [0, 1], [1, 0], [0, -1]])
    pix_per_cell = 1  # misnomer...
    half_pix_per_cell = pix_per_cell / 2
    x_y_half = array([half_pix_per_cell, -half_pix_per_cell])
    row_col_to_xy = array([[0, -pix_per_cell], [pix_per_cell, 0]])
    grid = array(np.meshgrid(np.arange(10), -np.arange(10))) * array([[[pix_per_cell]], [[pix_per_cell]]])
    theta = np.linspace(0, np.pi / 2, 5)
    arc = array([np.cos(theta), np.sin(theta)]).T  # from [1,0] to [0,1]

    def __init__(self, env, gl="PILSVG", jupyter=False, agent_render_variant=AgentRenderVariant.ONE_STEP_BEHIND):
        self.env = env
        self.frame_nr = 0
        self.start_time = time.time()
        self.times_list = deque()

        self.agent_render_variant = agent_render_variant

        if gl == "PIL":
            self.gl = PILGL(env.width, env.height, jupyter)
        elif gl == "PILSVG":
            self.gl = PILSVG(env.width, env.height, jupyter)
        else:
            print("[", gl, "] not found, switch to PILSVG")
            self.gl = PILSVG(env.width, env.height, jupyter)

        self.new_rail = True
        self.update_background()

    def reset(self):
        """
        Resets the environment
        :return:
        """
        self.set_new_rail()
        self.frame_nr = 0
        self.start_time = time.time()
        self.times_list = deque()
        return

    def update_background(self):
        # create background map
        targets = {}
        for agent_idx, agent in enumerate(self.env.agents_static):
            if agent is None:
                continue
            targets[tuple(agent.target)] = agent_idx
        self.gl.build_background_map(targets)

    def resize(self):
        self.gl.resize(self.env)

    def set_new_rail(self):
        """ Tell the renderer that the rail has changed.
            eg when the rail has been regenerated, or updated in the editor.
        """
        self.new_rail = True

    def plot_agents(self, targets=True, selected_agent=None):
        color_map = self.gl.get_cmap('hsv',
                                     lut=max(len(self.env.agents), len(self.env.agents_static) + 1))

        for agent_idx, agent in enumerate(self.env.agents_static):
            if agent is None:
                continue
            color = color_map(agent_idx)
            self.plot_single_agent(agent.position, agent.direction, color, target=agent.target if targets else None,
                                   static=True, selected=agent_idx == selected_agent)

        for agent_idx, agent in enumerate(self.env.agents):
            if agent is None:
                continue
            color = color_map(agent_idx)
            self.plot_single_agent(agent.position, agent.direction, color, target=agent.target if targets else None)

    def get_transition_row_col(self, row_col_pos, direction, bgiTrans=False):
        """
        Get the available transitions for row_col_pos in direction direction,
        as row & col deltas.

        If bgiTrans is True, return a grid of indices of available transitions.

        eg for a cell row_col_pos = (4,5), in direction direction = 0 (N),
        where the available transitions are N and E, returns:
        [[-1,0], [0,1]] ie N=up one row, and E=right one col.
        and if bgiTrans is True, returns a tuple:
        (
            [[-1,0], [0,1]], # deltas as before
            [0, 1] #  available transition indices, ie N, E
        )
        """

        transitions = self.env.rail.get_transitions(*row_col_pos, direction)
        transition_list = np.where(transitions)[0]  # RC list of transitions

        # HACK: workaround dead-end transitions
        if len(transition_list) == 0:
            reverse_direciton = (direction + 2) % 4
            transitions = tuple(int(tmp_dir == reverse_direciton) for tmp_dir in range(4))
            transition_list = np.where(transitions)[0]  # RC list of transitions

        transition_grid = self.__class__.transitions_row_col[transition_list]

        if bgiTrans:
            return transition_grid, transition_list
        else:
            return transition_grid

    def plot_single_agent(self, position_row_col, direction, color="r", target=None, static=False, selected=False):
        """
        Plot a simple agent.
        Assumes a working graphics layer context (cf a MPL figure).
        """
        rt = self.__class__

        direction_row_col = rt.transitions_row_col[direction]  # agent direction in RC
        direction_xy = np.matmul(direction_row_col, rt.row_col_to_xy)  # agent direction in xy

        xyPos = np.matmul(position_row_col - direction_row_col / 2, rt.row_col_to_xy) + rt.x_y_half

        if static:
            color = self.gl.adapt_color(color, lighten=True)

        color = color

        self.gl.scatter(*xyPos, color=color, layer=1, marker="o", s=100)  # agent location
        xy_dir_line = array([xyPos, xyPos + direction_xy / 2]).T  # line for agent orient.
        self.gl.plot(*xy_dir_line, color=color, layer=1, lw=5, ms=0, alpha=0.6)
        if selected:
            self._draw_square(xyPos, 1, color)

        if target is not None:
            target_row_col = array(target)
            target_xy = np.matmul(target_row_col, rt.row_col_to_xy) + rt.x_y_half
            self._draw_square(target_xy, 1 / 3, color, layer=1)

    def plot_transition(self, position_row_col, transition_row_col, color="r", depth=None):
        """
        plot the transitions in transition_row_col at position position_row_col.
        transition_row_col is a 2d numpy array containing a list of RC transitions,
        eg [[-1,0], [0,1]] means N, E.

        """

        rt = self.__class__
        position_xy = np.matmul(position_row_col, rt.row_col_to_xy) + rt.x_y_half
        transition_xy = position_xy + np.matmul(transition_row_col, rt.row_col_to_xy / 2.4)
        self.gl.scatter(*transition_xy.T, color=color, marker="o", s=50, alpha=0.2)
        if depth is not None:
            for x, y in transition_xy:
                self.gl.text(x, y, depth)

    def draw_transition(self,
                        line,
                        center,
                        rotation,
                        dead_end=False,
                        curves=False,
                        color="gray",
                        arrow=True,
                        spacing=0.1):
        """
        gLine is a numpy 2d array of points,
        in the plotting space / coords.
        eg:
        [[0,.5],[1,0.2]] means a line
        from x=0, y=0.5
        to   x=1, y=0.2
        """

        if not curves and not dead_end:
            # just a straigt line, no curve nor dead_end included in this basic rail element
            self.gl.plot(
                [line[0][0], line[1][0]],  # x
                [line[0][1], line[1][1]],  # y
                color=color
            )
        else:
            # it was not a simple line to draw: the rail has a curve or dead_end included.
            rt = self.__class__
            straight = rotation in [0, 2]
            dx, dy = np.squeeze(np.diff(line, axis=0)) * spacing / 2

            if straight:

                if color == "auto":
                    if dx > 0 or dy > 0:
                        color = "C1"  # N or E
                    else:
                        color = "C2"  # S or W

                if dead_end:
                    line_xy = array([
                        line[1] + [dy, dx],
                        center,
                        line[1] - [dy, dx],
                    ])
                    self.gl.plot(*line_xy.T, color=color)
                else:
                    line_xy = line + [-dy, dx]
                    self.gl.plot(*line_xy.T, color=color)

                    if arrow:
                        middle_xy = np.sum(line_xy * [[1 / 4], [3 / 4]], axis=0)

                        arrow_xy = array([
                            middle_xy + [-dx - dy, +dx - dy],
                            middle_xy,
                            middle_xy + [-dx + dy, -dx - dy]])
                        self.gl.plot(*arrow_xy.T, color=color)

            else:

                middle_xy = np.mean(line, axis=0)
                dxy = middle_xy - center
                corner = middle_xy + dxy
                if rotation == 1:
                    arc_factor = 1 - spacing
                    color_auto = "C1"
                else:
                    arc_factor = 1 + spacing
                    color_auto = "C2"
                dxy2 = (center - corner) * arc_factor  # for scaling the arc

                if color == "auto":
                    color = color_auto

                self.gl.plot(*(rt.arc * dxy2 + corner).T, color=color)

                if arrow:
                    dx, dy = np.squeeze(np.diff(line, axis=0)) / 20
                    iArc = int(len(rt.arc) / 2)
                    middle_xy = corner + rt.arc[iArc] * dxy2
                    arrow_xy = array([
                        middle_xy + [-dx - dy, +dx - dy],
                        middle_xy,
                        middle_xy + [-dx + dy, -dx - dy]])
                    self.gl.plot(*arrow_xy.T, color=color)

    def render_observation(self, agent_handles, observation_dict):
        """
        Render the extent of the observation of each agent. All cells that appear in the agent
        observation will be highlighted.
        :param agent_handles: List of agent indices to adapt color and get correct observation
        :param observation_dict: dictionary containing sets of cells of the agent observation

        """
        rt = self.__class__
        for agent in agent_handles:
            color = self.gl.get_agent_color(agent)
            for visited_cell in observation_dict[agent]:
                cell_coord = array(visited_cell[:2])
                cell_coord_trans = np.matmul(cell_coord, rt.row_col_to_xy) + rt.x_y_half
                self._draw_square(cell_coord_trans, 1 / (agent + 1.1), color, layer=1, opacity=100)

    def render_prediction(self, agent_handles, prediction_dict):
        """
        Render the extent of the observation of each agent. All cells that appear in the agent
        observation will be highlighted.
        :param agent_handles: List of agent indices to adapt color and get correct observation
        :param observation_dict: dictionary containing sets of cells of the agent observation

        """
        rt = self.__class__
        for agent in agent_handles:
            color = self.gl.get_agent_color(agent)
            for visited_cell in prediction_dict[agent]:
                cell_coord = array(visited_cell[:2])
                if type(self.gl) is PILSVG:
                    # TODO : Track highlighting (Adrian)
                    r = cell_coord[0]
                    c = cell_coord[1]
                    transitions = self.env.rail.grid[r, c]
                    self.gl.set_predicion_path_at(r, c, transitions, agent_rail_color=color)
                else:
                    cell_coord_trans = np.matmul(cell_coord, rt.row_col_to_xy) + rt.x_y_half
                    self._draw_square(cell_coord_trans, 1 / (agent + 1.1), color, layer=1, opacity=100)

    def render_rail(self, spacing=False, rail_color="gray", curves=True, arrows=False):

        cell_size = 1  # TODO: remove cell_size
        env = self.env

        # Draw cells grid
        grid_color = [0.95, 0.95, 0.95]
        for row in range(env.height + 1):
            self.gl.plot([0, (env.width + 1) * cell_size],
                         [-row * cell_size, -row * cell_size],
                         color=grid_color, linewidth=2)
        for col in range(env.width + 1):
            self.gl.plot([col * cell_size, col * cell_size],
                         [0, -(env.height + 1) * cell_size],
                         color=grid_color, linewidth=2)

        # Draw each cell independently
        for row in range(env.height):
            for col in range(env.width):

                # bounding box of the grid cell
                x0 = cell_size * col  # left
                x1 = cell_size * (col + 1)  # right
                y0 = cell_size * -row  # top
                y1 = cell_size * -(row + 1)  # bottom

                # centres of cell edges
                coords = [
                    ((x0 + x1) / 2.0, y0),  # N middle top
                    (x1, (y0 + y1) / 2.0),  # E middle right
                    ((x0 + x1) / 2.0, y1),  # S middle bottom
                    (x0, (y0 + y1) / 2.0)  # W middle left
                ]

                # cell centre
                center_xy = array([x0, y1]) + cell_size / 2

                # cell transition values
                cell = env.rail.get_full_transitions(row, col)

                cell_valid = env.rail.cell_neighbours_valid((row, col), check_this_cell=True)

                # Special Case 7, with a single bit; terminate at center
                nbits = 0
                tmp = cell

                while tmp > 0:
                    nbits += (tmp & 1)
                    tmp = tmp >> 1

                # as above - move the from coord to the centre
                # it's a dead env.
                is_dead_end = nbits == 1

                if not cell_valid:
                    self.gl.scatter(*center_xy, color="r", s=30)

                for orientation in range(4):  # ori is where we're heading
                    from_ori = (orientation + 2) % 4  # 0123=NESW -> 2301=SWNE
                    from_xy = coords[from_ori]

                    moves = env.rail.get_transitions(row, col, orientation)

                    for to_ori in range(4):
                        to_xy = coords[to_ori]
                        rotation = (to_ori - from_ori) % 4
                        if (moves[to_ori]):  # if we have this transition
                            self.draw_transition(
                                array([from_xy, to_xy]), center_xy,
                                rotation, dead_end=is_dead_end, curves=curves and not is_dead_end, spacing=spacing,
                                color=rail_color)

    def render_env(self,
                   show=False,  # whether to call matplotlib show() or equivalent after completion
                   agents=True,  # whether to include agents
                   show_observations=True,  # whether to include observations
                   show_predictions=False,  # whether to include predictions
                   frames=False,  # frame counter to show (intended since invocation)
                   episode=None,  # int episode number to show
                   step=None,  # int step number to show in image
                   selected_agent=None):  # indicate which agent is "selected" in the editor
        """ Draw the environment using the GraphicsLayer this RenderTool was created with.
            (Use show=False from a Jupyter notebook with %matplotlib inline)
        """
        if type(self.gl) is PILSVG:
            self.render_env_svg(show=show,
                                show_observations=show_observations,
                                show_predictions=show_predictions,
                                selected_agent=selected_agent
                                )
        else:
            self.render_env_pil(show=show,
                                agents=agents,
                                show_observations=show_observations,
                                show_predictions=show_predictions,
                                frames=frames,
                                episode=episode,
                                step=step,
                                selected_agent=selected_agent
                                )

    def _draw_square(self, center, size, color, opacity=255, layer=0):
        x0 = center[0] - size / 2
        x1 = center[0] + size / 2
        y0 = center[1] - size / 2
        y1 = center[1] + size / 2
        self.gl.plot([x0, x1, x1, x0, x0], [y0, y0, y1, y1, y0], color=color, layer=layer, opacity=opacity)

    def get_image(self):
        return self.gl.get_image()

    def render_env_pil(self,
                       show=False,  # whether to call matplotlib show() or equivalent after completion
                       # use false when calling from Jupyter.  (and matplotlib no longer supported!)
                       agents=True,  # whether to include agents
                       show_observations=True,  # whether to include observations
                       show_predictions=False,  # whether to include predictions
                       frames=False,  # frame counter to show (intended since invocation)
                       episode=None,  # int episode number to show
                       step=None,  # int step number to show in image
                       selected_agent=None  # indicate which agent is "selected" in the editor
                       ):

        if type(self.gl) is PILGL:
            self.gl.begin_frame()

        env = self.env

        self.render_rail()

        # Draw each agent + its orientation + its target
        if agents:
            self.plot_agents(targets=True, selected_agent=selected_agent)
        if show_observations:
            self.render_observation(range(env.get_num_agents()), env.dev_obs_dict)
        if show_predictions and len(env.dev_pred_dict) > 0:
            self.render_prediction(range(env.get_num_agents()), env.dev_pred_dict)
        # Draw some textual information like fps
        text_y = [-0.3, -0.6, -0.9]
        if frames:
            self.gl.text(0.1, text_y[2], "Frame:{:}".format(self.frame_nr))
        self.frame_nr += 1

        if episode is not None:
            self.gl.text(0.1, text_y[1], "Ep:{}".format(episode))

        if step is not None:
            self.gl.text(0.1, text_y[0], "Step:{}".format(step))

        time_now = time.time()
        self.gl.text(2, text_y[2], "elapsed:{:.2f}s".format(time_now - self.start_time))
        self.times_list.append(time_now)
        if len(self.times_list) > 20:
            self.times_list.popleft()
        if len(self.times_list) > 1:
            rFps = (len(self.times_list) - 1) / (self.times_list[-1] - self.times_list[0])
            self.gl.text(2, text_y[1], "fps:{:.2f}".format(rFps))

        self.gl.prettify2(env.width, env.height, self.pix_per_cell)

        # TODO: for MPL, we don't want to call clf (called by endframe)
        # if not show:

        if show and type(self.gl) is PILGL:
            self.gl.show()

        self.gl.pause(0.00001)

        return

    def render_env_svg(
        self, show=False, show_observations=True, show_predictions=False, selected_agent=None
    ):
        """
        Renders the environment with SVG support (nice image)
        """

        env = self.env

        self.gl.begin_frame()

        if self.new_rail:
            self.new_rail = False
            self.gl.clear_rails()

            # store the targets
            targets = {}
            selected = {}
            for agent_idx, agent in enumerate(self.env.agents_static):
                if agent is None:
                    continue
                targets[tuple(agent.target)] = agent_idx
                selected[tuple(agent.target)] = (agent_idx == selected_agent)

            # Draw each cell independently
            for r in range(env.height):
                for c in range(env.width):
                    transitions = env.rail.grid[r, c]
                    if (r, c) in targets:
                        target = targets[(r, c)]
                        is_selected = selected[(r, c)]
                    else:
                        target = None
                        is_selected = False

                    self.gl.set_rail_at(r, c, transitions, target=target, is_selected=is_selected,
                                        rail_grid=env.rail.grid)

            self.gl.build_background_map(targets)

        for agent_idx, agent in enumerate(self.env.agents):

            if agent is None:
                continue

            if self.agent_render_variant == AgentRenderVariant.BOX_ONLY:
                self.gl.set_cell_occupied(agent_idx, *(agent.position))
            elif self.agent_render_variant == AgentRenderVariant.ONE_STEP_BEHIND or \
                self.agent_render_variant == AgentRenderVariant.ONE_STEP_BEHIND_AND_BOX:  # noqa: E125
                if agent.old_position is not None:
                    position = agent.old_position
                    direction = agent.direction
                    old_direction = agent.old_direction
                else:
                    position = agent.position
                    direction = agent.direction
                    old_direction = agent.direction

                # set_agent_at uses the agent index for the color
                if self.agent_render_variant == AgentRenderVariant.ONE_STEP_BEHIND_AND_BOX:
                    self.gl.set_cell_occupied(agent_idx, *(agent.position))
                self.gl.set_agent_at(agent_idx, *position, old_direction, direction, selected_agent == agent_idx)
            else:
                position = agent.position
                direction = agent.direction
                for possible_directions in range(4):
                    # Is a transition along movement `desired_movement_from_new_cell' to the current cell possible?
                    isValid = env.rail.get_transition((*agent.position, agent.direction), possible_directions)
                    if isValid:
                        direction = possible_directions

                        # set_agent_at uses the agent index for the color
                        self.gl.set_agent_at(agent_idx, *position, agent.direction, direction,
                                             selected_agent == agent_idx)

                # set_agent_at uses the agent index for the color
                if self.agent_render_variant == AgentRenderVariant.AGENT_SHOWS_OPTIONS_AND_BOX:
                    self.gl.set_cell_occupied(agent_idx, *(agent.position))
                self.gl.set_agent_at(agent_idx, *position, agent.direction, direction, selected_agent == agent_idx)

        if show_observations:
            self.render_observation(range(env.get_num_agents()), env.dev_obs_dict)
        if show_predictions and len(env.dev_pred_dict) > 0:
            self.render_prediction(range(env.get_num_agents()), env.dev_pred_dict)
        if show:
            self.gl.show()
        for i in range(3):
            self.gl.process_events()

        self.frame_nr += 1
        return

    def close_window(self):
        self.gl.close_window()



"""
The env module defines the base Environment class.
The base Environment class is adapted from rllib.env.MultiAgentEnv
(https://github.com/ray-project/ray).
"""


class Environment:
    """
    Base interface for multi-agent environments in Flatland.

    Derived environments should implement the following attributes:
        action_space: tuple with the dimensions of the actions to be passed to the step method
        observation_space: tuple with the dimensions of the observations returned by reset and step

    Agents are identified by agent ids (handles).
    Examples:
        >>> obs = env.reset()
        >>> print(obs)
        {
            "train_0": [2.4, 1.6],
            "train_1": [3.4, -3.2],
        }
        >>> obs, rewards, dones, infos = env.step(
            action_dict={
                "train_0": 1, "train_1": 0})
        >>> print(rewards)
        {
            "train_0": 3,
            "train_1": -1,
        }
        >>> print(dones)
        {
            "train_0": False,    # train_0 is still running
            "train_1": True,     # train_1 is done
            "__all__": False,    # the env is not done
        }h
        >>> print(infos)
        {
            "train_0": {},  # info for train_0
            "train_1": {},  # info for train_1
        }
    """

    def __init__(self):
        self.action_space = ()
        self.observation_space = ()
        pass

    def reset(self):
        """
        Resets the env and returns observations from agents in the environment.

        Returns:
        obs : dict
            New observations for each agent.
        """
        raise NotImplementedError()

    def step(self, action_dict):
        """
        Environment step.

        Performs an environment step with simultaneous execution of actions for
        agents in action_dict. Returns observations for the agents.
        The returns are dicts mapping from agent_id strings to values.

        Parameters
        -------
        action_dict : dict
            Dictionary of actions to execute, indexed by agent id.

        Returns
        -------
        obs : dict
            New observations for each ready agent.
        rewards: dict
            Reward values for each ready agent.
        dones : dict
            Done values for each ready agent. The special key "__all__"
            (required) is used to indicate env termination.
        infos : dict
            Optional info values for each agent id.
        """
        raise NotImplementedError()

    def get_agent_handles(self):
        """
        Returns a list of agents' handles to be used as keys in the step()
        function.
        """
        raise NotImplementedError()
